from typing import Union
import logging

# noinspection PyUnresolvedReferences
import lab.torch
from algebra import Element
from matrix import AbstractMatrix
from plum import Dispatcher
from torch import Tensor

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover
from .random import Random

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

# Save original methods.
__mul__tensor = Tensor.__mul__
__imul__tensor = Tensor.__imul__
__add__tensor = Tensor.__add__
__iadd__tensor = Tensor.__iadd__


@_dispatch
def __mul__(self: Tensor, other):
    return __mul__tensor(self, other)


@_dispatch
def __mul__(self: Tensor, other: Union[Element, Random, AbstractMatrix]):
    return other.__rmul__(self)


@_dispatch
def __imul__(self: Tensor, other):
    return __imul__tensor(self, other)


@_dispatch
def __imul__(self: Tensor, other: Union[Element, Random, AbstractMatrix]):
    return other.__rmul__(self)


@_dispatch
def __add__(self: Tensor, other):
    return __add__tensor(self, other)


@_dispatch
def __add__(self: Tensor, other: Union[Element, Random, AbstractMatrix]):
    return other.__radd__(self)


@_dispatch
def __iadd__(self: Tensor, other):
    return __iadd__tensor(self, other)


@_dispatch
def __iadd__(self: Tensor, other: Union[Element, Random, AbstractMatrix]):
    return other.__radd__(self)


# Assign dispatchable methods.
Tensor.__mul__ = __mul__
Tensor.__imul__ = __imul__
Tensor.__add__ = __add__
Tensor.__iadd__ = __iadd__



import logging
from typing import Union

# noinspection PyUnresolvedReferences
import lab.autograd as B
from algebra import Element
from autograd.numpy.numpy_boxes import ArrayBox
from matrix import AbstractMatrix
from plum import Dispatcher

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover
from .random import Random

__all__ = []

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

# Save original methods.
__mul__array_box = ArrayBox.__mul__
__add__array_box = ArrayBox.__add__


@_dispatch
def __mul__(self: ArrayBox, other):
    return __mul__array_box(self, other)


@_dispatch
def __mul__(self: ArrayBox, other: Union[Element, Random, AbstractMatrix]):
    return other.__rmul__(self)


@_dispatch
def __add__(self: ArrayBox, other):
    return __add__array_box(self, other)


@_dispatch
def __add__(self: ArrayBox, other: Union[Element, Random, AbstractMatrix]):
    return other.__radd__(self)


# Assign dispatchable methods.
ArrayBox.__mul__ = __mul__
ArrayBox.__add__ = __add__



from .util import num_elements

__all__ = ["Input", "MultiInput", "Unique", "WeightedUnique"]


class Input:
    """Input type.

    Args:
        *xs (tensor): Inputs to type.
    """

    def __init__(self, *xs):
        self._xs = xs

    def get(self):
        """Get the wrapped input."""
        return self._xs[0] if len(self._xs) == 1 else self._xs


@num_elements.dispatch
def num_elements(x: Input):
    return sum(map(num_elements, x.get()))


class MultiInput(Input):
    """Multiple inputs."""

    def get(self):
        return self._xs


class Unique(Input):
    """One cannot learn about noise at these inputs."""


class WeightedUnique(Unique):
    """One cannot learn about noise at these inputs.

    Args:
        x (tensor): Input.
        w (tensor): Weights.
    """

    def __init__(self, x, w):
        self.w = w
        Unique.__init__(self, x)



from numbers import Number

import lab.tensorflow as B
import numpy as np
import pytest
import tensorflow as tf
from plum import Dispatcher

from stheno.input import Input, MultiInput
from stheno.kernel import EQ
from stheno.mean import TensorProductMean, ZeroMean, Mean, OneMean, PosteriorMean
from stheno.measure import FDD
from .util import approx

_dispatch = Dispatcher()


@_dispatch
def f1(x: Number):
    return np.array([[x ** 2]])


@_dispatch
def f1(x):
    return B.sum(x ** 2, axis=1)[:, None]


@_dispatch
def f2(x: Number):
    return np.array([[x ** 3]])


@_dispatch
def f2(x):
    return B.sum(x ** 3, axis=1)[:, None]


def test_corner_cases():
    with pytest.raises(RuntimeError):
        Mean()(1.0)


@pytest.mark.parametrize("x", [B.randn(10), Input(B.randn(10)), FDD(None, B.randn(10))])
def test_construction(x):
    m = TensorProductMean(lambda y: y ** 2)
    m(x)

    # Test `MultiInput` construction.
    approx(m(MultiInput(x, x)), B.concat(m(x), m(x), axis=0))


def test_basic_arithmetic():
    m1 = TensorProductMean(f1)
    m2 = TensorProductMean(f2)
    m3 = ZeroMean()

    x1 = B.randn(10, 2)
    x2 = B.randn()

    approx((m1 * m2)(x1), m1(x1) * m2(x1))
    approx((m1 * m2)(x2), m1(x2) * m2(x2))
    approx((m1 + m3)(x1), m1(x1) + m3(x1))
    approx((m1 + m3)(x2), m1(x2) + m3(x2))
    approx((5.0 * m1)(x1), 5.0 * m1(x1))
    approx((5.0 * m1)(x2), 5.0 * m1(x2))
    approx((5.0 + m1)(x1), 5.0 + m1(x1))
    approx((5.0 + m1)(x2), 5.0 + m1(x2))


def test_sum():
    m1 = TensorProductMean(f1)
    m2 = TensorProductMean(f2)

    # Test equality.
    assert m1 + m2 == m1 + m2
    assert m1 + m2 == m2 + m1
    assert m1 + m2 != ZeroMean() + m2
    assert m1 + m2 != m1 + ZeroMean()


def test_product():
    m1 = TensorProductMean(f1)
    m2 = TensorProductMean(f2)

    # Test equality.
    assert m1 * m2 == m1 * m2
    assert m1 * m2 == m2 * m1
    assert m1 * m2 != m2 * m2
    assert m1 * m2 != m1 * m1


def test_posterior_mean():
    z = B.linspace(0, 1, 10)
    pcm = PosteriorMean(
        TensorProductMean(lambda x: x),
        TensorProductMean(lambda x: x ** 2),
        EQ(),
        z,
        2 * EQ()(z),
        B.randn(10),
    )

    # Check name.
    assert str(pcm) == "PosteriorMean()"

    # Check that the mean computes.
    pcm(z)


def test_tensor_product():
    m1 = 5 * OneMean() + (lambda x: x ** 2)
    m2 = (lambda x: x ** 2) + 5 * OneMean()
    m3 = (lambda x: x ** 2) + ZeroMean()
    m4 = ZeroMean() + (lambda x: x ** 2)

    x = B.randn(10, 1)
    assert np.allclose(m1(x), 5 + x ** 2)
    assert np.allclose(m2(x), 5 + x ** 2)
    assert np.allclose(m3(x), x ** 2)
    assert np.allclose(m4(x), x ** 2)

    def my_function(x):
        pass

    assert str(TensorProductMean(my_function)) == "my_function"


def test_derivative():
    m = TensorProductMean(lambda x: x ** 2)
    m2 = TensorProductMean(lambda x: x ** 3)

    x = B.randn(tf.float64, 10, 1)
    approx(m.diff(0)(x), 2 * x)
    approx(m2.diff(0)(x), 3 * x ** 2)


def test_selected_mean():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = B.randn(10, 3)
    approx(m.select([1, 2])(x), m(x[:, [1, 2]]))


def test_shifting():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = B.randn(10, 3)
    approx(m.shift(5)(x), m(x - 5))


def test_stretching():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = B.randn(10, 3)
    approx(m.stretch(5)(x), m(x / 5))


def test_input_transform():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = B.randn(10, 3)
    approx(m.transform(lambda x: x - 5)(x), m(x - 5))


import jax
import jax.numpy as jnp
import lab as B
import pytest
import tensorflow as tf
import torch
from algebra import Element
import autograd
from lab import torch as B
from matrix import AbstractMatrix, Dense

from stheno.torch import EQ


@pytest.mark.parametrize("grad", [autograd.grad, lambda f: jax.jit(jax.grad(f))])
def test_boxing_objective(grad):
    k = EQ()
    objs = []

    def objective(x):
        x = x + 2
        x = x * 2
        objs.append(x[0] + k)
        objs.append(x[0] + Dense(k(x)))
        objs.append(x[0] * k)
        objs.append(x[0] * Dense(k(x)))
        return B.sum(x)

    grad(objective)(B.randn(10))

    for obj in objs:
        assert isinstance(obj, (Element, AbstractMatrix))


@pytest.mark.parametrize(
    "generate, Numeric",
    [
        (lambda *shape: B.randn(torch.float64, *shape), B.TorchNumeric),
        (lambda *shape: B.randn(tf.float64, *shape), B.TFNumeric),
        (lambda *shape: tf.Variable(B.randn(tf.float64, *shape)), B.TFNumeric),
    ],
)
def test_boxing_direct(generate, Numeric):
    k = EQ()
    num = generate()
    x = generate(10)

    # Test addition and multiplication.
    assert isinstance(num + k, Element)
    assert isinstance(num + Dense(k(x)), AbstractMatrix)
    assert isinstance(num + num, Numeric)
    assert isinstance(num * k, Element)
    assert isinstance(num * Dense(k(x)), AbstractMatrix)
    assert isinstance(num * num, Numeric)


@pytest.mark.parametrize(
    "generate, Numeric",
    [
        (lambda *shape: B.randn(torch.float64, *shape), B.TorchNumeric),
        (lambda *shape: B.randn(tf.float64, *shape), B.TFNumeric),
    ],
)
def test_boxing_inplace(generate, Numeric):
    k = EQ()
    x = generate(10)

    # Test in-place addition.
    num = generate()
    num += generate()
    assert isinstance(num, Numeric)

    num += k
    assert isinstance(num, Element)

    num = generate()
    num += Dense(k(x))
    assert isinstance(num, AbstractMatrix)

    # Test in-place multiplication.
    num = generate()
    num *= generate()
    assert isinstance(num, Numeric)

    num *= k
    assert isinstance(num, Element)

    num = generate()
    num *= Dense(k(x))
    assert isinstance(num, AbstractMatrix)



import pytest

from stheno.lazy import LazyMatrix, LazyVector


def test_indexing():
    v = LazyVector()
    m = LazyMatrix()

    v[1] = 1
    assert v[1] == 1

    # Test diagonal indexing and setting.
    m[1] = 1
    m[2, 2] = 2
    assert m[1, 1] == 1
    assert m[2] == 2

    # Test resolving of indices.
    class A:
        pass

    a1, a2 = A(), A()
    m[id(a1), a2] = (1, 2)
    m[id(a1)] = 1
    m[a2] = 2

    assert m[a1, a1] == 1
    assert m[a1, a2] == (1, 2)
    assert m[a2, a2] == 2


class Number:
    def __init__(self, x):
        self.x = x


def test_lazy_vector_building():
    v = LazyVector()

    # Test universal building.
    v.add_rule(set(range(3)), lambda x: Number(x))

    for i in range(3):
        assert v[i].x == i

    # Test bounds checking.
    with pytest.raises(RuntimeError):
        v[-1]
    with pytest.raises(RuntimeError):
        v[3]

    # Stack another universal building rule.
    v.add_rule(set(range(4)), lambda x: Number(3 + x))

    for i in range(3):
        assert v[i].x == i

    assert v[3].x == 3 + 3

    # Test bounds checking.
    with pytest.raises(RuntimeError):
        v[-1]
    with pytest.raises(RuntimeError):
        v[4]

    # Now try all rules at once and mix up access order.
    v2 = LazyVector()
    v2.add_rule(set(range(3)), lambda x: Number(x))
    v2.add_rule(set(range(4)), lambda x: Number(3 + x))

    assert v2[3].x == 3 + 3

    for i in range(3):
        assert v2[i].x == i

    # Test bounds checking.
    with pytest.raises(RuntimeError):
        v2[-1]
    with pytest.raises(RuntimeError):
        v2[4]


def test_lazy_vector_mutation_protection():
    inds = {1}
    v = LazyVector()
    v.add_rule(inds, lambda i: 1)
    inds.add(2)
    assert v[1] == 1
    with pytest.raises(RuntimeError):
        v[2]


def test_lazy_matrix_building():
    m = LazyMatrix()

    def check_bounds(bounds):
        for bound in bounds:
            with pytest.raises(RuntimeError):
                m[bound]

    # Test universal building.
    m.add_rule(set(range(3)), lambda x, y: Number(x * y))

    # Check content.
    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j

    # Check bounds.
    check_bounds([(-1, 0), (0, -1), (-1, -1), (4, 3), (3, 4), (4, 4)])

    # Test left and right building rules.
    m.add_left_rule(3, set(range(4)), lambda y: Number(3 + y))
    m.add_right_rule(3, set(range(4)), lambda y: Number(3 + y))

    # Check content.
    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j
    for i in range(4):
        assert m[i, 3].x == i + 3
        assert m[3, i].x == 3 + i
    assert m[3, 3].x == 3 + 3

    # Check bounds.
    check_bounds([(-1, 0), (0, -1), (-1, -1), (5, 4), (4, 5), (5, 5)])

    # Stack more left and right building rules.
    m.add_right_rule(4, set(range(5)), lambda x: Number(x ** 2 + 4 ** 2))
    m.add_left_rule(4, set(range(5)), lambda x: Number(x ** 2 + 4 ** 2))

    # Check content.
    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j
    for i in range(4):
        assert m[i, 3].x == i + 3
        assert m[3, i].x == 3 + i
    assert m[3, 3].x == 3 + 3
    for i in range(5):
        assert m[i, 4].x == i ** 2 + 4 ** 2
        assert m[4, i].x == 4 ** 2 + i ** 2
    assert m[4, 4].x == 4 ** 2 + 4 ** 2

    # Check bounds.
    check_bounds([(-1, 0), (0, -1), (-1, -1), (6, 5), (5, 6), (6, 6)])

    # Now try all rules at once and mix up access order.
    m2 = LazyMatrix()
    m2.add_rule(set(range(3)), lambda x, y: Number(x * y))
    m2.add_left_rule(3, set(range(4)), lambda y: Number(3 + y))
    m2.add_right_rule(3, set(range(4)), lambda y: Number(3 + y))
    m2.add_right_rule(4, set(range(5)), lambda x: Number(x ** 2 + 4 ** 2))
    m2.add_left_rule(4, set(range(5)), lambda x: Number(x ** 2 + 4 ** 2))

    # Check content.
    for i in range(5):
        assert m2[i, 4].x == i ** 2 + 4 ** 2
        assert m2[4, i].x == 4 ** 2 + i ** 2
    assert m2[4, 4].x == 4 ** 2 + 4 ** 2
    for i in range(4):
        assert m2[i, 3].x == i + 3
        assert m2[3, i].x == 3 + i
    assert m2[3, 3].x == 3 + 3
    for i in range(3):
        for j in range(3):
            assert m2[i, j].x == i * j

    # Check bounds.
    check_bounds([(-1, 0), (0, -1), (-1, -1), (6, 5), (5, 6), (6, 6)])


def test_lazy_matrix_mutation_protection():
    def check_access(m):
        assert m[1, 1] == 1
        with pytest.raises(RuntimeError):
            m[1, 2]
        with pytest.raises(RuntimeError):
            m[2, 1]
        with pytest.raises(RuntimeError):
            m[2, 2]

    # Check universal rule.
    inds = {1}
    m = LazyMatrix()
    m.add_rule(inds, lambda i, j: 1)
    inds.add(2)
    check_access(m)

    # Check left rule.
    inds = {1}
    m = LazyMatrix()
    m.add_left_rule(1, inds, lambda j: 1)
    inds.add(2)
    check_access(m)

    # Check right rule.
    inds = {1}
    m = LazyMatrix()
    m.add_right_rule(1, inds, lambda i: 1)
    inds.add(2)
    check_access(m)



import logging

import algebra as algebra
from lab import B
from matrix import AbstractMatrix
from plum import Dispatcher, convert, Union

from . import PromisedFDD as FDD
from .input import Input, MultiInput
from .kernel import unwrap
from .util import num_elements, uprank

__all__ = ["TensorProductMean", "DerivativeMean"]

_dispatch = Dispatcher()

log = logging.getLogger(__name__)


class Mean(algebra.Function):
    """Mean function.

    Means can be added and multiplied.
    """

    @_dispatch
    def __call__(self, x):
        """Construct the mean for a design matrix.

        Args:
            x (input): Points to construct the mean at.

        Returns:
            tensor: Mean vector as a rank 2 column vector.
        """
        raise RuntimeError(f'For mean {self}, could not resolve argument "{x}".')

    @_dispatch
    def __call__(self, x: Union[Input, FDD]):
        return self(unwrap(x))

    @_dispatch
    def __call__(self, x: MultiInput):
        return B.concat(*[self(xi) for xi in x.get()], axis=0)


# Register the algebra.
@algebra.get_algebra.dispatch
def get_algebra(a: Mean):
    return Mean


class SumMean(Mean, algebra.SumFunction):
    """Sum of two means."""

    @_dispatch
    def __call__(self, x):
        return B.add(self[0](x), self[1](x))


class ProductMean(Mean, algebra.ProductFunction):
    """Product of two means."""

    @_dispatch
    def __call__(self, x):
        return B.multiply(self[0](x), self[1](x))


class ScaledMean(Mean, algebra.ScaledFunction):
    """Scaled mean."""

    @_dispatch
    def __call__(self, x):
        return B.multiply(self.scale, self[0](x))


class StretchedMean(Mean, algebra.StretchedFunction):
    """Stretched mean."""

    @_dispatch
    def __call__(self, x):
        return self[0](B.divide(x, self.stretches[0]))


class ShiftedMean(Mean, algebra.ShiftedFunction):
    """Shifted mean."""

    @_dispatch
    def __call__(self, x):
        return self[0](B.subtract(x, self.shifts[0]))


class SelectedMean(Mean, algebra.SelectedFunction):
    """Mean with particular input dimensions selected."""

    @_dispatch
    @uprank
    def __call__(self, x):
        return self[0](B.take(x, self.dims[0], axis=1))


class InputTransformedMean(Mean, algebra.InputTransformedFunction):
    """Input-transformed mean."""

    @_dispatch
    def __call__(self, x):
        return self[0](uprank(self.fs[0](uprank(x))))


class OneMean(Mean, algebra.OneFunction):
    """Constant mean of `1`."""

    @_dispatch
    def __call__(self, x: B.Numeric):
        return B.ones(B.dtype(x), num_elements(x), 1)


class ZeroMean(Mean, algebra.ZeroFunction):
    """Constant mean of `0`."""

    @_dispatch
    def __call__(self, x: B.Numeric):
        return B.zeros(B.dtype(x), num_elements(x), 1)


class TensorProductMean(Mean, algebra.TensorProductFunction):
    @_dispatch
    def __call__(self, x: B.Numeric):
        return uprank(self.fs[0](uprank(x)))


class DerivativeMean(Mean, algebra.DerivativeFunction):
    """Derivative of mean."""

    @_dispatch
    @uprank
    def __call__(self, x: B.Numeric):
        import tensorflow as tf

        i = self.derivs[0]
        with tf.GradientTape() as t:
            xi = x[:, i : i + 1]
            t.watch(xi)
            x = B.concat(x[:, :i], xi, x[:, i + 1 :], axis=1)
            out = B.dense(self[0](x))
            return t.gradient(out, xi, unconnected_gradients="zero")


class PosteriorMean(Mean):
    """Posterior mean.

    Args:
        m_i (:class:`.mean.Mean`): Mean of process corresponding to the input.
        m_z (:class:`.mean.Mean`): Mean of process corresponding to the data.
        k_zi (:class:`.kernel.Kernel`): Kernel between processes corresponding to the
            data and the input respectively.
        z (input): Locations of data.
        K_z (matrix): Kernel matrix of data.
        y (tensor): Observations to condition on.
    """

    def __init__(self, m_i, m_z, k_zi, z, K_z, y):
        self.m_i = m_i
        self.m_z = m_z
        self.k_zi = k_zi
        self.z = z
        self.K_z = convert(K_z, AbstractMatrix)
        self.y = uprank(y)

    @_dispatch
    def __call__(self, x):
        diff = B.subtract(self.y, self.m_z(self.z))
        return B.add(self.m_i(x), B.iqf(self.K_z, self.k_zi(self.z, x), diff))



from functools import wraps
from types import FunctionType

from lab import B
from plum import Dispatcher

__all__ = ["num_elements", "uprank"]

_dispatch = Dispatcher()


@_dispatch
def num_elements(x):
    """Determine the number of elements in an input.

    Deals with scalars, vectors, and matrices.

    Args:
        x (tensor): Input.

    Returns:
        int: Number of elements.
    """
    shape = B.shape(x)
    if shape == ():
        return 1
    else:
        return shape[0]


@_dispatch
def uprank(x):
    """Ensure that the rank of `x` is 2.

    Args:
        x (tensor): Tensor to uprank.

    Returns:
        tensor: `x` with rank at least 2.
    """
    # Simply return non-numerical inputs.
    if not isinstance(x, B.Numeric):
        return x
    else:
        return B.uprank(x)


@_dispatch
def uprank(f: FunctionType):
    """A decorator to ensure that the rank of the arguments is two."""

    @wraps(f)
    def wrapped_f(*args):
        return f(*[uprank(x) for x in args])

    return wrapped_f


import numpy as np
import pytest
import tensorflow as tf
from lab import B
from plum import NotFoundLookupError

from stheno.input import Input
from stheno.kernel import (
    Linear,
    EQ,
    Delta,
    Exp,
    ZeroKernel,
    OneKernel,
    ScaledKernel,
)
from stheno.mean import TensorProductMean, ZeroMean, ScaledMean, OneMean
from stheno.measure import Measure, GP, Obs, SparseObs, FDD, cross
from stheno.random import Normal
from .util import approx


def assert_equal_gps(p1, p2):
    assert p1.mean == p2.mean
    assert p1.kernel == p2.kernel


def assert_equal_normals(d1, d2):
    approx(d1.mean, d2.mean)
    approx(d1.var, d2.var)


def assert_equal_measures(fdds, post_ref, *posts):
    for post in posts:
        for fdd in fdds:
            assert_equal_normals(post_ref(fdd), post(fdd))


def test_corner_cases():
    p1 = GP(EQ())
    p2 = GP(EQ())
    x = B.randn(10, 2)

    # Test check for measure group.
    with pytest.raises(AssertionError):
        p1 + p2
    with pytest.raises(AssertionError):
        p1 * p2

    # Test incompatible operations.
    with pytest.raises(NotFoundLookupError):
        p1 + p1(x)
    with pytest.raises(NotFoundLookupError):
        p1 * p1(x)

    # Check display of GPs.
    assert str(GP()) == "GP()"
    assert str(GP(EQ())) == "GP(0, EQ())"

    # Check test for prior.
    with pytest.raises(RuntimeError):
        GP().measure


def test_construction():
    p = GP(EQ())

    x = B.randn(10, 1)

    p.mean(x)
    p.mean(Input(x))

    p.kernel(x)
    p.kernel(Input(x))
    p.kernel(x, x)
    p.kernel(Input(x), x)
    p.kernel(x, Input(x))
    p.kernel(Input(x), Input(x))

    p.kernel.elwise(x)
    p.kernel.elwise(Input(x))
    p.kernel.elwise(x, x)
    p.kernel.elwise(Input(x), x)
    p.kernel.elwise(x, Input(x))
    p.kernel.elwise(Input(x), Input(x))

    # Test resolution of kernel and mean.
    k = EQ()
    m = TensorProductMean(lambda x: x ** 2)

    assert isinstance(GP(k).mean, ZeroMean)
    assert isinstance(GP(5, k).mean, ScaledMean)
    assert isinstance(GP(1, k).mean, OneMean)
    assert isinstance(GP(0, k).mean, ZeroMean)
    assert isinstance(GP(m, k).mean, TensorProductMean)
    assert isinstance(GP(k).kernel, EQ)
    assert isinstance(GP(5).kernel, ScaledKernel)
    assert isinstance(GP(1).kernel, OneKernel)
    assert isinstance(GP(0).kernel, ZeroKernel)

    # Test construction of finite-dimensional distribution.
    d = GP(m, k)(x)
    approx(d.var, k(x))
    approx(d.mean, m(x))


def test_sum_other():
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())

    def five(y):
        return 5 * B.ones(B.shape(y)[0], 1)

    x = B.randn(5, 1)
    for p_sum in [
        # Add a numeric thing.
        p + 5.0,
        5.0 + p,
        p.measure.sum(GP(), p, 5.0),
        p.measure.sum(GP(), 5.0, p),
        # Add a function.
        p + five,
        five + p,
        p.measure.sum(GP(), p, five),
        p.measure.sum(GP(), five, p),
    ]:
        approx(p.mean(x) + 5.0, p_sum.mean(x))
        approx(p.mean(x) + 5.0, p_sum.mean(x))
        approx(p.kernel(x), p_sum.kernel(x))
        approx(p.kernel(x), p_sum.kernel(x))

    # Check that a `GP` cannot be summed with a `Normal`.
    with pytest.raises(NotFoundLookupError):
        p + Normal(np.eye(3))
    with pytest.raises(NotFoundLookupError):
        Normal(np.eye(3)) + p


def test_mul_other():
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())

    def five(y):
        return 5 * B.ones(B.shape(y)[0], 1)

    x = B.randn(5, 1)
    for p_mul in [
        # Multiply numeric thing.
        p * 5.0,
        5.0 * p,
        p.measure.mul(GP(), p, 5.0),
        p.measure.mul(GP(), 5.0, p),
        # Multiply with a function.
        p * five,
        five * p,
        p.measure.mul(GP(), p, five),
        p.measure.mul(GP(), five, p),
    ]:
        approx(5.0 * p.mean(x), p_mul.mean(x))
        approx(5.0 * p.mean(x), p_mul.mean(x))
        approx(25.0 * p.kernel(x), p_mul.kernel(x))
        approx(25.0 * p.kernel(x), p_mul.kernel(x))

    # Check that a `GP` cannot be multiplied with a `Normal`.
    with pytest.raises(NotFoundLookupError):
        p * Normal(np.eye(3))
    with pytest.raises(NotFoundLookupError):
        Normal(np.eye(3)) * p


def test_fdd():
    p = GP(EQ())

    # Test construction.
    fdd = p(1)
    assert isinstance(fdd, FDD)
    assert fdd.x is 1
    assert fdd.p is p

    class A:
        def __str__(self):
            return "str"

        def __repr__(self):
            return "repr"

    # Check representation.
    assert str(p(A())) == "FDD(GP(0, EQ()), str)"
    assert repr(p(A())) == "FDD(GP(0, EQ()), repr)"


def test_measure_groups():
    prior = Measure()
    f1 = GP(EQ(), measure=prior)
    f2 = GP(EQ(), measure=prior)

    assert f1._measures == f2._measures

    x = B.linspace(0, 5, 10)
    y = f1(x).sample()

    post = prior | (f1(x), y)

    assert f1._measures == f2._measures == [prior, post]

    # Further extend the prior.

    f_sum = f1 + f2
    assert f_sum._measures == [prior, post]

    f3 = GP(EQ(), measure=prior)
    f_sum = f1 + f3
    assert f3._measures == f_sum._measures == [prior]

    with pytest.raises(AssertionError):
        post(f1) + f3

    # Extend the posterior.

    f_sum = post(f1) + post(f2)
    assert f_sum._measures == [post]

    f3 = GP(EQ(), measure=post)
    f_sum = post(f1) + f3
    assert f3._measures == f_sum._measures == [post]

    with pytest.raises(AssertionError):
        f1 + f3


def test_shorthands():
    p = GP(EQ())
    # Test shorthands for stretching and selection.
    assert_equal_gps(p > 2, p.stretch(2))
    assert_equal_gps(p[0], p.select(0))


def test_stationarity():
    m = Measure()

    p1 = GP(EQ(), measure=m)
    p2 = GP(EQ().stretch(2), measure=m)
    p3 = GP(EQ().periodic(10), measure=m)

    p = p1 + 2 * p2

    assert p.stationary

    p = p3 + p

    assert p.stationary

    p = p + GP(Linear(), measure=m)

    assert not p.stationary


def test_marginals():
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())
    x = B.linspace(0, 5, 10)

    # Check that `marginals` outputs the right thing.
    mean, lower, upper = p(x).marginals()
    var = B.diag(p.kernel(x))
    approx(mean, p.mean(x)[:, 0])
    approx(lower, p.mean(x)[:, 0] - 1.96 * var ** 0.5)
    approx(upper, p.mean(x)[:, 0] + 1.96 * var ** 0.5)

    # Test correctness.
    y = p(x).sample()
    post = p.measure | (p(x), y)

    mean, lower, upper = post(p)(x).marginals()
    approx(mean, y[:, 0])
    approx(upper - lower, B.zeros(10), atol=1e-5)

    mean, lower, upper = post(p)(x + 100).marginals()
    approx(mean, p.mean(x + 100)[:, 0])
    approx(upper - lower, 3.92 * B.ones(10))


def test_conditioning():
    m = Measure()
    p1 = GP(EQ(), measure=m)
    p2 = GP(Exp(), measure=m)
    p_sum = p1 + p2

    # Sample some data to condition on.
    x1 = B.linspace(0, 2, 2)
    y1 = p1(x1).sample()
    x_sum = B.linspace(3, 5, 3)
    y_sum = p_sum(x_sum).sample()

    # Determine FDDs to check.
    x_check = B.linspace(0, 5, 5)
    fdds_check = [
        cross(p1, p2, p_sum)(x_check),
        p1(x_check),
        p2(x_check),
        p_sum(x_check),
    ]

    assert_equal_measures(
        fdds_check,
        m.condition(p_sum(x_sum), y_sum),
        m.condition((p_sum(x_sum), y_sum)),
        m | (p_sum(x_sum), y_sum),
        m | ((p_sum(x_sum), y_sum),),
        m | Obs(p_sum(x_sum), y_sum),
        m | Obs((p_sum(x_sum), y_sum)),
    )

    assert_equal_measures(
        fdds_check,
        m.condition((p1(x1), y1), (p_sum(x_sum), y_sum)),
        m | ((p1(x1), y1), (p_sum(x_sum), y_sum)),
        m | Obs((p1(x1), y1), (p_sum(x_sum), y_sum)),
    )


def test_conditioning_prior():
    p = GP(EQ())
    x = B.zeros(0, 1)
    y = B.zeros(0, 1)
    post = p.measure | (p(x), y)
    assert post(p).mean is p.mean
    assert post(p).kernel is p.kernel


def test_sparse_conditioning_and_elbo():
    m = Measure()
    p1 = GP(EQ(), measure=m)
    p2 = GP(Exp(), measure=m)
    e = GP(Delta(), measure=m)
    p_sum = p1 + p2

    # Sample some data to condition on.
    x1 = B.linspace(0, 2, 2)
    y1 = (p1 + e)(x1).sample()
    x_sum = B.linspace(3, 5, 3)
    y_sum = (p_sum + e)(x_sum).sample()

    # Determine FDDs to check.
    x_check = B.linspace(0, 5, 5)
    fdds_check = [
        cross(p1, p2, p_sum)(x_check),
        p1(x_check),
        p2(x_check),
        p_sum(x_check),
    ]

    # Check conditioning and ELBO on one data set.
    assert_equal_measures(
        fdds_check,
        m | ((p_sum + e)(x_sum), y_sum),
        m | SparseObs(p_sum(x_sum), e, p_sum(x_sum), y_sum),
        m | SparseObs((p_sum(x_sum),), e, p_sum(x_sum), y_sum),
        m | SparseObs((p_sum(x_sum), p1(x1)), e, p_sum(x_sum), y_sum),
        m | SparseObs(p_sum(x_sum), (e, p_sum(x_sum), y_sum)),
        m | SparseObs((p_sum(x_sum),), (e, p_sum(x_sum), y_sum)),
        m.condition(
            SparseObs(
                (p_sum(x_sum), p1(x1)),
                (e, p_sum(x_sum), y_sum),
            )
        ),
    )
    approx(
        m.logpdf(Obs((p_sum + e)(x_sum), y_sum)),
        SparseObs(p_sum(x_sum), (e, p_sum(x_sum), y_sum)).elbo(m),
    )

    # Check conditioning and ELBO on two data sets.
    assert_equal_measures(
        fdds_check,
        m | (((p_sum + e)(x_sum), y_sum), ((p1 + e)(x1), y1)),
        m.condition(
            SparseObs((p_sum(x_sum), p1(x1)), (e, p_sum(x_sum), y_sum), (e, p1(x1), y1))
        ),
    )
    approx(
        m.logpdf(Obs(((p_sum + e)(x_sum), y_sum), ((p1 + e)(x1), y1))),
        SparseObs(
            (p_sum(x_sum), p1(x1)), (e, p_sum(x_sum), y_sum), (e, p1(x1), y1)
        ).elbo(m),
    )

    # The following lose information, so check them separately.
    assert_equal_measures(
        fdds_check,
        m | SparseObs(p_sum(x_sum), (e, p_sum(x_sum), y_sum), (e, p1(x1), y1)),
        m | SparseObs((p_sum(x_sum),), (e, p_sum(x_sum), y_sum), (e, p1(x1), y1)),
    )

    # Test lazy computation.
    obs = SparseObs(p_sum(x_sum), e, p_sum(x_sum), y_sum)
    for name in ["K_z", "elbo", "mu", "A"]:
        approx(
            getattr(SparseObs(p_sum(x_sum), e, p_sum(x_sum), y_sum), name)(m),
            getattr(obs, name)(m),
        )

    # Test requirement that noise must be diagonal.
    with pytest.raises(RuntimeError):
        SparseObs(p_sum(x_sum), p_sum, p_sum(x_sum), y_sum).elbo(m)


def test_logpdf():
    m = Measure()
    p1 = GP(EQ(), measure=m)
    p2 = GP(Exp(), measure=m)
    e = GP(Delta(), measure=m)
    p3 = p1 + p2

    x1 = B.linspace(0, 2, 5)
    x2 = B.linspace(1, 3, 6)
    x3 = B.linspace(2, 4, 7)
    y1, y2, y3 = m.sample(p1(x1), p2(x2), p3(x3))

    # Test case that only one process is fed.
    approx(p1(x1).logpdf(y1), m.logpdf(p1(x1), y1))
    approx(p1(x1).logpdf(y1), m.logpdf((p1(x1), y1)))

    # Compute the logpdf with the product rule.
    d1 = m
    d2 = d1 | (p1(x1), y1)
    d3 = d2 | (p2(x2), y2)
    approx(
        d1(p1)(x1).logpdf(y1) + d2(p2)(x2).logpdf(y2) + d3(p3)(x3).logpdf(y3),
        m.logpdf((p1(x1), y1), (p2(x2), y2), (p3(x3), y3)),
    )

    # Check that `Measure.logpdf` allows `Obs` and `SparseObs`.
    obs = Obs(p3(x3), y3)
    approx(m.logpdf(obs), p3(x3).logpdf(y3))
    obs = SparseObs(p3(x3), e, p3(x3), y3)
    approx(m.logpdf(obs), (p3 + e)(x3).logpdf(y3))


def test_stretching():
    # Test construction:
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())
    assert str(p.stretch(1)) == "GP(<lambda> > 1, EQ() > 1)"

    # Test case:
    p = GP(EQ())
    p_stretched = p.stretch(5)

    x = B.linspace(0, 5, 10)
    y = p_stretched(x).sample()

    post = p.measure | (p_stretched(x), y)
    assert_equal_normals(post(p(x / 5)), post(p_stretched(x)))
    assert_equal_normals(post(p(x)), post(p_stretched(x * 5)))


def test_shifting():
    # Test construction:
    p = GP(TensorProductMean(lambda x: x ** 2), Linear())
    assert str(p.shift(1)) == "GP(<lambda> shift 1, Linear() shift 1)"

    # Test case:
    p = GP(EQ())
    p_shifted = p.shift(5)

    x = B.linspace(0, 5, 10)
    y = p_shifted(x).sample()

    post = p.measure | (p_shifted(x), y)
    assert_equal_normals(post(p(x - 5)), post(p_shifted(x)))
    assert_equal_normals(post(p(x)), post(p_shifted(x + 5)))


def test_input_transform():
    # Test construction:
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())
    assert (
        str(p.transform(lambda x: x))
        == "GP(<lambda> transform <lambda>, EQ() transform <lambda>)"
    )

    # Test case:
    p = GP(EQ())
    p_transformed = p.transform(lambda x: B.sqrt(x))

    x = B.linspace(0, 5, 10)
    y = p_transformed(x).sample()

    post = p.measure | (p_transformed(x), y)
    assert_equal_normals(post(p(B.sqrt(x))), post(p_transformed(x)))
    assert_equal_normals(post(p(x)), post(p_transformed(x * x)))


def test_selection():
    # Test construction:
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())
    assert str(p.select(1)) == "GP(<lambda> : [1], EQ() : [1])"
    assert str(p.select(1, 2)) == "GP(<lambda> : [1, 2], EQ() : [1, 2])"

    # Test case:
    p = GP(EQ())  # 1D
    p2 = p.select(0)  # 2D

    x = B.linspace(0, 5, 10)
    x21 = B.stack(x, B.randn(10), axis=1)
    x22 = B.stack(x, B.randn(10), axis=1)
    y = p2(x).sample()

    post = p.measure | (p2(x21), y)
    approx(post(p(x)).mean, y)
    assert_equal_normals(post(p(x)), post(p2(x21)))

    post = p.measure | (p2(x22), y)
    approx(post(p(x)).mean, y)
    assert_equal_normals(post(p(x)), post(p2(x22)))

    post = p.measure | (p(x), y)
    approx(post(p2(x21)).mean, y)
    approx(post(p2(x22)).mean, y)
    assert_equal_normals(post(p2(x21)), post(p(x)))
    assert_equal_normals(post(p2(x22)), post(p(x)))


def test_derivative():
    # Test construction:
    p = GP(TensorProductMean(lambda x: x ** 2), EQ())
    assert str(p.diff(1)) == "GP(d(1) <lambda>, d(1) EQ())"

    # Test case:
    p = GP(EQ())
    dp = p.diff()

    x = B.linspace(tf.float64, 0, 1, 100)
    y = 2 * x

    x_check = B.linspace(tf.float64, 0.2, 0.8, 100)

    # Test conditioning on function.
    post = p.measure | (p(x), y)
    approx(post(dp)(x_check).mean, 2 * B.ones(100, 1), atol=1e-4)

    # Test conditioning on derivative.
    zero = B.cast(tf.float64, 0)
    post = p.measure | ((p(zero), zero), (dp(x), y))
    approx(post(p)(x_check).mean, x_check[:, None] ** 2, atol=1e-4)


def test_multi_sample():
    m = Measure()
    p1 = GP(1, 0, measure=m)
    p2 = GP(2, 0, measure=m)
    p3 = GP(3, 0, measure=m)

    x1 = B.linspace(0, 1, 5)
    x2 = B.linspace(0, 1, 10)
    x3 = B.linspace(0, 1, 15)

    s1, s2, s3 = m.sample(p1(x1), p2(x2), p3(x3))

    assert B.shape(p1(x1).sample()) == s1.shape == (5, 1)
    assert B.shape(p2(x2).sample()) == s2.shape == (10, 1)
    assert B.shape(p3(x3).sample()) == s3.shape == (15, 1)

    approx(s1, 1 * B.ones(5, 1))
    approx(s2, 2 * B.ones(10, 1))
    approx(s3, 3 * B.ones(15, 1))


def test_approximate_multiplication():
    m = Measure()
    p1 = GP(20, EQ(), measure=m)
    p2 = GP(20, EQ(), measure=m)
    p_prod = p1 * p2

    # Sample functions.
    x = B.linspace(0, 10, 50)
    s1, s2 = m.sample(p1(x), p2(x))

    # Perform product.
    post = m | ((p1(x), s1), (p2(x), s2))
    approx(post(p_prod)(x).mean, s1 * s2, rtol=1e-2)

    # Perform division.
    cur_epsilon = B.epsilon
    B.epsilon = 1e-8
    post = m | ((p1(x), s1), (p_prod(x), s1 * s2))
    approx(post(p2)(x).mean, s2, rtol=1e-2)
    B.epsilon = cur_epsilon


def test_naming():
    m = Measure()

    p1 = GP(EQ(), 1, measure=m)
    p2 = GP(EQ(), 2, measure=m)

    # Test setting and getting names.
    p1.name = "name"

    assert m["name"] is p1
    assert p1.name == "name"
    assert m[p1] == "name"
    with pytest.raises(KeyError):
        m["other_name"]
    with pytest.raises(KeyError):
        m[p2]

    # Check that names can not be doubly assigned.
    def doubly_assign():
        p2.name = "name"

    with pytest.raises(RuntimeError):
        doubly_assign()

    # Move name to other GP.
    p1.name = "other_name"
    p2.name = "name"

    # Check that everything has been properly assigned.
    assert m["name"] is p2
    assert p2.name == "name"
    assert m[p2] == "name"
    assert m["other_name"] is p1
    assert p1.name == "other_name"
    assert m[p1] == "other_name"

    # Test giving a name to the constructor.
    p3 = GP(EQ(), name="yet_another_name", measure=m)
    assert m["yet_another_name"] is p3
    assert p3.name == "yet_another_name"
    assert m[p3] == "yet_another_name"


def test_formatting():
    p = 2 * GP(1, EQ(), measure=Measure())
    assert str(p.display(lambda x: x ** 2)) == "GP(4 * 1, 16 * EQ())"


def test_case_summation_with_itself():
    p = GP(EQ())
    p_many = p + p + p + p + p

    x = B.linspace(0, 10, 5)
    approx(p_many(x).var, 25 * p(x).var)
    approx(p_many(x).mean, B.zeros(5, 1))

    y = B.randn(5, 1)
    post = p.measure | (p(x), y)
    approx(post(p_many)(x).mean, 5 * y)


def test_case_additive_model():
    m = Measure()
    p1 = GP(EQ(), measure=m)
    p2 = GP(EQ(), measure=m)
    p_sum = p1 + p2

    x = B.linspace(0, 5, 10)
    y1 = p1(x).sample()
    y2 = p2(x).sample()

    # First, test independence:
    assert m.kernels[p2, p1] == ZeroKernel()
    assert m.kernels[p1, p2] == ZeroKernel()

    # Now run through some test cases:
    post = (m | (p1(x), y1)) | (p2(x), y2)
    approx(post(p_sum)(x).mean, y1 + y2)

    post = (m | (p2(x), y2)) | (p1(x), y1)
    approx(post(p_sum)(x).mean, y1 + y2)

    post = (m | (p1(x), y1)) | (p_sum(x), y1 + y2)
    approx(post(p2)(x).mean, y2)

    post = (m | (p_sum(x), y1 + y2)) | (p1(x), y1)
    approx(post(p2)(x).mean, y2)

    post = (m | (p2(x), y2)) | (p_sum(x), y1 + y2)
    approx(post(p1)(x).mean, y1)

    post = (m | (p_sum(x), y1 + y2)) | (p2(x), y2)
    approx(post(p1)(x).mean, y1)


def test_case_fd_derivative():
    x = B.linspace(0, 10, 50)
    y = np.sin(x)

    p = GP(0.7 * EQ().stretch(1.0))
    dp = (p.shift(-1e-3) - p.shift(1e-3)) / 2e-3

    post = p.measure | (p(x), y)
    approx(post(dp)(x).mean, np.cos(x)[:, None], atol=1e-4)


def test_case_reflection():
    p = GP(EQ())
    p2 = 5 - p

    x = B.linspace(0, 5, 10)
    y = p(x).sample()

    post = p.measure | (p(x), y)
    approx(post(p2)(x).mean, 5 - y)

    post = p.measure | (p2(x), 5 - y)
    approx(post(p)(x).mean, y)


def test_case_negation():
    p = GP(EQ())
    p2 = -p

    x = B.linspace(0, 5, 10)
    y = p(x).sample()

    post = p.measure | (p(x), y)
    approx(post(p2)(x).mean, -y)

    post = p.measure | (p2(x), -y)
    approx(post(p)(x).mean, y)


def test_case_approximate_derivative():
    p = GP(EQ().stretch(1.0))
    dp = p.diff_approx()

    x = B.linspace(0, 1, 100)
    y = 2 * x

    x_check = B.linspace(0.2, 0.8, 100)

    # Test conditioning on function.
    post = p.measure | (p(x), y)
    approx(post(dp)(x_check).mean, 2 * B.ones(100, 1), atol=1e-3)

    # Test conditioning on derivative.
    orig_epsilon = B.epsilon
    B.epsilon = 1e-10
    post = p.measure | ((p(0), 0), (dp(x), y))
    approx(post(p)(x_check).mean, x_check[:, None] ** 2, atol=1e-3)
    B.epsilon = orig_epsilon


def test_case_blr():
    m = Measure()
    x = B.linspace(0, 10, 100)

    slope = GP(1, measure=m)
    intercept = GP(1, measure=m)
    f = slope * (lambda x: x) + intercept
    y = f + 1e-2 * GP(Delta(), measure=m)

    # Sample observations, true slope, and intercept.
    y_obs, true_slope, true_intercept = m.sample(y(x), slope(0), intercept(0))

    # Predict.
    post = m | (y(x), y_obs)
    approx(post(slope)(0).mean, true_slope, atol=5e-2)
    approx(post(intercept)(0).mean, true_intercept, atol=5e-2)



import logging

import algebra
import numpy as np
from algebra.util import identical, to_tensor
from lab import B
from matrix import Dense, LowRank, Diagonal, Zero, Constant, AbstractMatrix
from plum import Dispatcher, convert, Union

from . import PromisedFDD as FDD
from .input import Input, Unique, WeightedUnique, MultiInput
from .util import num_elements, uprank

__all__ = [
    "Kernel",
    "OneKernel",
    "ZeroKernel",
    "ScaledKernel",
    "EQ",
    "RQ",
    "Matern12",
    "Exp",
    "Matern32",
    "Matern52",
    "Delta",
    "FixedDelta",
    "Linear",
    "DerivativeKernel",
    "DecayingKernel",
    "LogKernel",
]

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


@_dispatch
def expand(xs: Union[tuple, list]):
    """Expand a sequence to the same element repeated twice if there is only one
    element.

    Args:
        xs (tuple or list): Sequence to expand.

    Returns:
        tuple or list: `xs * 2` or `xs`.
    """
    return xs * 2 if len(xs) == 1 else xs


@_dispatch
def unwrap(x: Input):
    """Unwrap a wrapped input.

    Args:
        x (input): Wrapped input.

    Returns:
        tensor: Unwrapped input.
    """
    return x.get()


@_dispatch
def unwrap(fdd: FDD):
    return fdd.x


class Kernel(algebra.Function):
    """Kernel function.

    Kernels can be added and multiplied.
    """

    @_dispatch
    def __call__(self, x, y):
        """Construct the kernel matrix between all `x` and `y`.

        Args:
            x (input): First argument.
            y (input, optional): Second argument. Defaults to first
                argument.

        Returns:
            matrix: Kernel matrix.
        """
        raise RuntimeError(
            f'For kernel "{self}", could not resolve arguments "{x}" and "{y}".'
        )

    @_dispatch
    def __call__(self, x):
        return self(x, x)

    @_dispatch
    def __call__(self, x: Union[Input, FDD], y: Union[Input, FDD]):
        return self(unwrap(x), unwrap(y))

    @_dispatch
    def __call__(self, x: Union[Input, FDD], y):
        return self(unwrap(x), y)

    @_dispatch
    def __call__(self, x, y: Union[Input, FDD]):
        return self(x, unwrap(y))

    @_dispatch(precedence=1)
    def __call__(self, x: MultiInput, y):
        return self(x, MultiInput(y))

    @_dispatch(precedence=1)
    def __call__(self, x, y: MultiInput):
        return self(MultiInput(x), y)

    @_dispatch
    def __call__(self, x: MultiInput, y: MultiInput):
        return B.block(*[[self(xi, yi) for yi in y.get()] for xi in x.get()])

    @_dispatch
    def elwise(self, x, y):
        """Construct the kernel vector `x` and `y` element-wise.

        Args:
            x (input): First argument.
            y (input, optional): Second argument. Defaults to first
                argument.

        Returns:
            tensor: Kernel vector as a rank 2 column vector.
        """
        # TODO: Throw warning.
        return B.expand_dims(B.diag(self(x, y)), axis=1)

    @_dispatch
    def elwise(self, x):
        return self.elwise(x, x)

    @_dispatch
    def elwise(self, x: Union[Input, FDD], y: Union[Input, FDD]):
        return self.elwise(unwrap(x), unwrap(y))

    @_dispatch
    def elwise(self, x: Union[Input, FDD], y):
        return self.elwise(unwrap(x), y)

    @_dispatch
    def elwise(self, x, y: Union[Input, FDD]):
        return self.elwise(x, unwrap(y))

    @_dispatch(precedence=1)
    def elwise(self, x: MultiInput, y):
        raise ValueError("Unclear combination of arguments given to `Kernel.elwise`.")

    @_dispatch(precedence=1)
    def elwise(self, x, y: MultiInput):
        raise ValueError("Unclear combination of arguments given to `Kernel.elwise`.")

    @_dispatch
    def elwise(self, x: MultiInput, y: MultiInput):
        if len(x.get()) != len(y.get()):
            raise ValueError(
                "Kernel.elwise must be called with similarly sized MultiInputs."
            )
        return B.concat(
            *[self.elwise(xi, yi) for xi, yi in zip(x.get(), y.get())], axis=0
        )

    def periodic(self, period=1):
        """Map to a periodic space.

        Args:
            period (tensor, optional): Period. Defaults to `1`.

        Returns:
            :class:`.kernel.Kernel`: Periodic version of the kernel.
        """
        return periodicise(self, period)

    @property
    def stationary(self):
        """Stationarity of the kernel."""
        try:
            return self._stationary_cache
        except AttributeError:
            self._stationary_cache = self._stationary
            return self._stationary_cache

    @property
    def _stationary(self):
        return False


# Register the algebra.
@algebra.get_algebra.dispatch
def get_algebra(a: Kernel):
    return Kernel


class OneKernel(Kernel, algebra.OneFunction):
    """Constant kernel of `1`."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Constant(B.one(x), num_elements(x), num_elements(y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return B.ones(B.dtype(x), num_elements(x), 1)

    @property
    def _stationary(self):
        return True


class ZeroKernel(Kernel, algebra.ZeroFunction):
    """Constant kernel of `0`."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return B.zeros(B.dtype(x), num_elements(x), 1)

    @property
    def _stationary(self):
        return True


class ScaledKernel(Kernel, algebra.ScaledFunction):
    """Scaled kernel."""

    @_dispatch
    def __call__(self, x, y):
        return self._compute(self[0](x, y))

    @_dispatch
    def elwise(self, x, y):
        return self._compute(self[0].elwise(x, y))

    def _compute(self, K):
        return B.multiply(self.scale, K)

    @property
    def _stationary(self):
        return self[0].stationary


class SumKernel(Kernel, algebra.SumFunction):
    """Sum of kernels."""

    @_dispatch
    def __call__(self, x, y):
        return B.add(self[0](x, y), self[1](x, y))

    @_dispatch
    def elwise(self, x, y):
        return B.add(self[0].elwise(x, y), self[1].elwise(x, y))

    @property
    def _stationary(self):
        return self[0].stationary and self[1].stationary


class ProductKernel(Kernel, algebra.ProductFunction):
    """Product of two kernels."""

    @_dispatch
    def __call__(self, x, y):
        return B.multiply(self[0](x, y), self[1](x, y))

    @_dispatch
    def elwise(self, x, y):
        return B.multiply(self[0].elwise(x, y), self[1].elwise(x, y))

    @property
    def _stationary(self):
        return self[0].stationary and self[1].stationary


class StretchedKernel(Kernel, algebra.StretchedFunction):
    """Stretched kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return self[0](*self._compute(x, y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        stretches1, stretches2 = expand(self.stretches)
        return B.divide(x, stretches1), B.divide(y, stretches2)

    @property
    def _stationary(self):
        if len(self.stretches) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @_dispatch
    def __eq__(self, other: "StretchedKernel"):
        identical_stretches = identical(expand(self.stretches), expand(other.stretches))
        return self[0] == other[0] and identical_stretches


class ShiftedKernel(Kernel, algebra.ShiftedFunction):
    """Shifted kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return self[0](*self._compute(x, y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        shifts1, shifts2 = expand(self.shifts)
        return B.subtract(x, shifts1), B.subtract(y, shifts2)

    @property
    def _stationary(self):
        if len(self.shifts) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @_dispatch
    def __eq__(self, other: "ShiftedKernel"):
        identical_shifts = identical(expand(self.shifts), expand(other.shifts))
        return self[0] == other[0] and identical_shifts


class SelectedKernel(Kernel, algebra.SelectedFunction):
    """Kernel with particular input dimensions selected."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return self[0](*self._compute(x, y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self[0].elwise(*self._compute(x, y))

    @uprank
    def _compute(self, x, y):
        dims1, dims2 = expand(self.dims)
        x = x if dims1 is None else B.take(x, dims1, axis=1)
        y = y if dims2 is None else B.take(y, dims2, axis=1)
        return x, y

    @property
    def _stationary(self):
        if len(self.dims) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @_dispatch
    def __eq__(self, other: "SelectedKernel"):
        return self[0] == other[0] and identical(expand(self.dims), expand(other.dims))


class InputTransformedKernel(Kernel, algebra.InputTransformedFunction):
    """Input-transformed kernel."""

    @_dispatch
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        f1, f2 = expand(self.fs)
        x = x if f1 is None else f1(uprank(x))
        y = y if f2 is None else f2(uprank(y))
        return x, y

    @_dispatch
    def __eq__(self, other: "InputTransformedKernel"):
        return self[0] == other[0] and identical(expand(self.fs), expand(other.fs))


class PeriodicKernel(Kernel, algebra.WrappedFunction):
    """Periodic kernel.

    Args:
        k (:class:`.kernel.Kernel`): Kernel to make periodic.
        scale (tensor): Period.
    """

    def __init__(self, k, period):
        algebra.WrappedFunction.__init__(self, k)
        self.period = to_tensor(period)

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return self[0](*self._compute(x, y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        @uprank
        def feature_map(z):
            z = B.divide(B.multiply(B.multiply(z, 2), B.pi), self.period)
            return B.concat(B.sin(z), B.cos(z), axis=1)

        return feature_map(x), feature_map(y)

    @property
    def _stationary(self):
        return self[0].stationary

    def render_wrap(self, e, formatter):
        return f"{e} per {formatter(self.period)}"

    @_dispatch
    def __eq__(self, other: "PeriodicKernel"):
        return self[0] == other[0] and identical(self.period, other.period)


class EQ(Kernel):
    """Exponentiated quadratic kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        return B.exp(-0.5 * dists2)

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "EQ"):
        return True


class RQ(Kernel):
    """Rational quadratic kernel.

    Args:
        alpha (scalar): Shape of the prior over length scales. Determines the
            weight of the tails of the kernel. Must be positive.
    """

    def __init__(self, alpha):
        self.alpha = alpha

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        return (1 + 0.5 * dists2 / self.alpha) ** (-self.alpha)

    def render(self, formatter):
        return f"RQ({formatter(self.alpha)})"

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "RQ"):
        return B.all(self.alpha == other.alpha)


class Exp(Kernel):
    """Exponential kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Dense(B.exp(-B.pw_dists(x, y)))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return B.exp(-B.ew_dists(x, y))

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "Exp"):
        return True


Matern12 = Exp  #: Alias for the exponential kernel.


class Matern32(Kernel):
    """Matern--3/2 kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Dense(self._compute(B.pw_dists(x, y)))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self._compute(B.ew_dists(x, y))

    def _compute(self, dists):
        r = 3 ** 0.5 * dists
        return (1 + r) * B.exp(-r)

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "Matern32"):
        return True


class Matern52(Kernel):
    """Matern--5/2 kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        return Dense(self._compute(B.pw_dists(x, y)))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return self._compute(B.ew_dists(x, y))

    def _compute(self, dists):
        r1 = 5 ** 0.5 * dists
        r2 = 5 * dists ** 2 / 3
        return (1 + r1 + r2) * B.exp(-r1)

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "Matern52"):
        return True


class Delta(Kernel):
    """Kronecker delta kernel.

    Args:
        epsilon (float, optional): Tolerance for equality in squared distance.
            Defaults to `1e-10`.
    """

    def __init__(self, epsilon=1e-10):
        self.epsilon = epsilon

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        if x is y:
            return B.fill_diag(B.one(x), num_elements(x))
        else:
            return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch
    def __call__(self, x: Unique, y: Unique):
        x, y = x.get(), y.get()
        if x is y:
            return B.fill_diag(B.one(x), num_elements(x))
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def __call__(self, x: WeightedUnique, y: WeightedUnique):
        w_x, w_y = x.w, y.w
        x, y = x.get(), y.get()
        if x is y:
            return Diagonal(1 / w_x)
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def __call__(self, x: Unique, y):
        x = x.get()
        return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def __call__(self, x, y: Unique):
        y = y.get()
        return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def elwise(self, x: Unique, y: Unique):
        x, y = x.get(), y.get()
        if x is y:
            return B.ones(B.dtype(x), num_elements(x), 1)
        else:
            return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch
    def elwise(self, x: WeightedUnique, y: WeightedUnique):
        w_x, w_y = x.w, y.w
        x, y = x.get(), y.get()
        if x is y:
            return uprank(1 / w_x)
        else:
            return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch
    def elwise(self, x: Unique, y):
        x = x.get()
        return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch
    def elwise(self, x, y: Unique):
        return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        if x is y:
            return B.ones(B.dtype(x), num_elements(x), 1)
        else:
            return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        dtype = B.dtype(dists2)
        return B.cast(dtype, B.lt(dists2, B.cast(dtype, self.epsilon)))

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "Delta"):
        return self.epsilon == other.epsilon


class FixedDelta(Kernel):
    """Kronecker delta kernel that produces a diagonal matrix with given
    noises if and only if the inputs are identical and of the right shape.

    Args:
        noises (vector): Noises.
    """

    def __init__(self, noises):
        self.noises = noises

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        if x is y and num_elements(x) == B.shape(self.noises)[0]:
            return Diagonal(self.noises)
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        if x is y and num_elements(x) == B.shape(self.noises)[0]:
            return uprank(self.noises)
        else:
            return Zero(B.dtype(x), num_elements(x), 1)

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "FixedDelta"):
        return B.all(self.noises == other.noises)


class Linear(Kernel):
    """Linear kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        if x is y:
            return LowRank(uprank(x))
        else:
            return LowRank(left=uprank(x), right=uprank(y))

    @_dispatch
    @uprank
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return B.expand_dims(B.sum(B.multiply(x, y), axis=1), axis=1)

    @property
    def _stationary(self):
        return False

    @_dispatch
    def __eq__(self, other: "Linear"):
        return True


class DecayingKernel(Kernel):
    """Decaying kernel.

    Args:
        alpha (tensor): Shape of the gamma distribution governing the
            distribution of decaying exponentials.
        beta (tensor): Rate of the gamma distribution governing the
            distribution of decaying exponentials.
    """

    def __init__(self, alpha, beta):
        self.alpha = alpha
        self.beta = beta

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        pw_sums_raised = B.power(B.pw_sums(B.add(x, self.beta), y), self.alpha)
        return Dense(B.divide(self._compute_beta_raised(), pw_sums_raised))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        return B.divide(
            self._compute_beta_raised(),
            B.power(B.ew_sums(B.add(x, self.beta), y), self.alpha),
        )

    def _compute_beta_raised(self):
        beta_norm = B.sqrt(
            B.maximum(B.sum(B.power(self.beta, 2)), B.cast(B.dtype(self.beta), 1e-30))
        )
        return B.power(beta_norm, self.alpha)

    def render(self, formatter):
        return f"DecayingKernel({formatter(self.alpha)}, {formatter(self.beta)})"

    @_dispatch
    def __eq__(self, other: "DecayingKernel"):
        return B.all(self.alpha == other.alpha) and B.all(self.beta == other.beta)


class LogKernel(Kernel):
    """Logarithm kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        dists = B.maximum(B.pw_dists(x, y), 1e-10)
        return Dense(B.divide(B.log(dists + 1), dists))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        dists = B.maximum(B.ew_dists(x, y), 1e-10)
        return B.divide(B.log(dists + 1), dists)

    def render(self, formatter):
        return "LogKernel()"

    @property
    def _stationary(self):
        return True

    @_dispatch
    def __eq__(self, other: "LogKernel"):
        return True


class PosteriorKernel(Kernel):
    """Posterior kernel.

    Args:
        k_ij (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the left input and the right input respectively.
        k_zi (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the left input respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the right input respectively.
        z (input): Locations of data.
        K_z (matrix): Kernel matrix of data.
    """

    def __init__(self, k_ij, k_zi, k_zj, z, K_z):
        self.k_ij = k_ij
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.K_z = convert(K_z, AbstractMatrix)

    @_dispatch
    def __call__(self, x, y):
        return B.subtract(
            self.k_ij(x, y), B.iqf(self.K_z, self.k_zi(self.z, x), self.k_zj(self.z, y))
        )

    @_dispatch
    def elwise(self, x, y):
        iqf_diag = B.iqf_diag(self.K_z, self.k_zi(self.z, x), self.k_zj(self.z, y))
        return B.subtract(self.k_ij.elwise(x, y), B.expand_dims(iqf_diag, axis=1))


class CorrectiveKernel(Kernel):
    """Kernel that adds the corrective variance in sparse conditioning.

    Args:
        k_zi (:class:`.kernel.Kernel`): Kernel between the processes corresponding to
            the left input and the inducing points respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between the processes corresponding to
            the right input and the inducing points respectively.
        z (input): Locations of the inducing points.
        A (tensor): Corrective matrix.
        L (tensor): Kernel matrix of the inducing points.
    """

    def __init__(self, k_zi, k_zj, z, A, K_z):
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.A = A
        self.L = B.cholesky(convert(K_z, AbstractMatrix))

    @_dispatch
    def __call__(self, x, y):
        return B.iqf(
            self.A,
            B.solve(self.L, self.k_zi(self.z, x)),
            B.solve(self.L, self.k_zj(self.z, y)),
        )

    @_dispatch
    def elwise(self, x, y):
        return B.iqf_diag(
            self.A,
            B.solve(self.L, self.k_zi(self.z, x)),
            B.solve(self.L, self.k_zj(self.z, y)),
        )[:, None]


def dkx(k_elwise, i):
    """Construct the derivative of a kernel with respect to its first
    argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Derivative of the kernel with respect to its first argument.
    """

    @uprank
    def _dkx(x, y):
        import tensorflow as tf

        with tf.GradientTape() as t:
            # Get the numbers of inputs.
            nx = num_elements(x)
            ny = num_elements(y)

            # Copy the input `ny` times to efficiently compute many derivatives.
            xis = tf.identity_n([x[:, i : i + 1]] * ny)
            t.watch(xis)

            # Tile inputs for batched computation.
            x = B.tile(x, ny, 1)
            y = B.reshape(B.tile(y, 1, nx), ny * nx, -1)

            # Insert tracked dimension, which is different for every tile.
            xi = B.concat(*xis, axis=0)
            x = B.concat(x[:, :i], xi, x[:, i + 1 :], axis=1)

            # Perform the derivative computation.
            out = B.dense(k_elwise(x, y))
            grads = t.gradient(out, xis, unconnected_gradients="zero")
            return B.concat(*grads, axis=1)

    return _dkx


def dkx_elwise(k_elwise, i):
    """Construct the element-wise derivative of a kernel with respect to
    its first argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Element-wise derivative of the kernel with respect to its
            first argument.
    """

    @uprank
    def _dkx_elwise(x, y):
        import tensorflow as tf

        with tf.GradientTape() as t:
            xi = x[:, i : i + 1]
            t.watch(xi)
            x = B.concat(x[:, :i], xi, x[:, i + 1 :], axis=1)
            out = B.dense(k_elwise(x, y))
            return t.gradient(out, xi, unconnected_gradients="zero")

    return _dkx_elwise


def dky(k_elwise, i):
    """Construct the derivative of a kernel with respect to its second
    argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Derivative of the kernel with respect to its second argument.
    """

    @uprank
    def _dky(x, y):
        import tensorflow as tf

        with tf.GradientTape() as t:
            # Get the numbers of inputs.
            nx = num_elements(x)
            ny = num_elements(y)

            # Copy the input `nx` times to efficiently compute many derivatives.
            yis = tf.identity_n([y[:, i : i + 1]] * nx)
            t.watch(yis)

            # Tile inputs for batched computation.
            x = B.reshape(B.tile(x, 1, ny), nx * ny, -1)
            y = B.tile(y, nx, 1)

            # Insert tracked dimension, which is different for every tile.
            yi = B.concat(*yis, axis=0)
            y = B.concat(y[:, :i], yi, y[:, i + 1 :], axis=1)

            # Perform the derivative computation.
            out = B.dense(k_elwise(x, y))
            grads = t.gradient(out, yis, unconnected_gradients="zero")
            return B.transpose(B.concat(*grads, axis=1))

    return _dky


def dky_elwise(k_elwise, i):
    """Construct the element-wise derivative of a kernel with respect to
    its second argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Element-wise derivative of the kernel with respect to its
            second argument.
    """

    @uprank
    def _dky_elwise(x, y):
        import tensorflow as tf

        with tf.GradientTape() as t:
            yi = y[:, i : i + 1]
            t.watch(yi)
            y = B.concat(y[:, :i], yi, y[:, i + 1 :], axis=1)
            out = B.dense(k_elwise(x, y))
            return t.gradient(out, yi, unconnected_gradients="zero")

    return _dky_elwise


def perturb(x):
    """Slightly perturb a tensor.

    Args:
        x (tensor): Tensor to perturb.

    Returns:
        tensor: `x`, but perturbed.
    """
    dtype = convert(B.dtype(x), B.NPDType)
    if dtype == np.float64:
        return 1e-20 + x * (1 + 1e-14)
    elif dtype == np.float32:
        return 1e-20 + x * (1 + 1e-7)
    else:
        raise ValueError(f"Cannot perturb a tensor of data type {B.dtype(x)}.")


class DerivativeKernel(Kernel, algebra.DerivativeFunction):
    """Derivative of kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        i, j = expand(self.derivs)
        k = self[0]

        # Prevent that `x` equals `y` to stabilise nested gradients.
        y = perturb(y)

        if i is not None and j is not None:
            # Derivative with respect to both `x` and `y`.
            return Dense(dky(dkx_elwise(k.elwise, i), j)(x, y))

        elif i is not None and j is None:
            # Derivative with respect to `x`.
            return Dense(dkx(k.elwise, i)(x, y))

        elif i is None and j is not None:
            # Derivative with respect to `y`.
            return Dense(dky(k.elwise, j)(x, y))

        else:
            raise RuntimeError("No derivative specified.")

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        i, j = expand(self.derivs)
        k = self[0]

        # Prevent that `x` equals `y` to stabilise nested gradients.
        y = perturb(y)

        if i is not None and j is not None:
            # Derivative with respect to both `x` and `y`.
            return dky_elwise(dkx_elwise(k.elwise, i), j)(x, y)

        elif i is not None and j is None:
            # Derivative with respect to `x`.
            return dkx_elwise(k.elwise, i)(x, y)

        elif i is None and j is not None:
            # Derivative with respect to `y`.
            return dky_elwise(k.elwise, j)(x, y)

        else:
            raise RuntimeError("No derivative specified.")

    @property
    def _stationary(self):
        # NOTE: In the one-dimensional case, if derivatives with respect to both
        # arguments are taken, then the result is in fact stationary.
        return False

    @_dispatch
    def __eq__(self, other: "DerivativeKernel"):
        identical_derivs = identical(expand(self.derivs), expand(other.derivs))
        return self[0] == other[0] and identical_derivs


class TensorProductKernel(Kernel, algebra.TensorProductFunction):
    """Tensor product kernel."""

    @_dispatch
    def __call__(self, x: B.Numeric, y: B.Numeric):
        f1, f2 = expand(self.fs)
        if x is y and f1 is f2:
            return LowRank(uprank(f1(uprank(x))))
        else:
            return LowRank(left=uprank(f1(uprank(x))), right=uprank(f2(uprank(y))))

    @_dispatch
    def elwise(self, x: B.Numeric, y: B.Numeric):
        f1, f2 = expand(self.fs)
        if x is y and f1 is f2:
            return B.power(uprank(f1(uprank(x))), 2)
        else:
            return B.multiply(uprank(f1(uprank(x))), uprank(f2(uprank(y))))

    @_dispatch
    def __eq__(self, other: "TensorProductKernel"):
        return identical(expand(self.fs), expand(other.fs))


class ReversedKernel(Kernel, algebra.ReversedFunction):
    """Reversed kernel.

    Evaluates with its arguments reversed.
    """

    @_dispatch
    def __call__(self, x, y):
        return B.transpose(self[0](y, x))

    @_dispatch
    def elwise(self, x, y):
        return self[0].elwise(y, x)

    @property
    def _stationary(self):
        return self[0].stationary


# Periodicise kernels.


@_dispatch
def periodicise(a: Kernel, b):
    return PeriodicKernel(a, b)


@_dispatch
def periodicise(a: ZeroKernel, b):
    return a


# Make shifting synergise with stationary kernels.


@algebra.shift.dispatch
def shift(a: Kernel, *shifts):
    if a.stationary and len(shifts) == 1:
        return a
    else:
        return ShiftedKernel(a, *shifts)



import jax

# noinspection PyUnresolvedReferences
import lab.jax as B

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover

# We need `float64`s for accurate kernels.
jax.config.update("jax_enable_x64", True)



from typing import Union

import logging
import abc

from plum import Dispatcher

__all__ = []

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


@_dispatch
def _resolve_index(key):
    return id(key)


@_dispatch
def _resolve_index(i: int):
    return i


@_dispatch
def _resolve_index(x: Union[tuple, reversed]):
    return tuple(_resolve_index(key) for key in x)


class LazyTensor:
    """A lazy tensor that indexes by the identity of its indices.

    Args:
        rank (int): Rank of the tensor.
    """

    def __init__(self, rank):
        self._rank = rank
        self._store = {}

    @_dispatch
    def _expand_key(self, key: tuple):
        # Nothing to do. The key needs to be a tuple.
        return key

    @_dispatch
    def _expand_key(self, key):
        return (key,) * self._rank

    def __setitem__(self, key, value):
        return self._set(_resolve_index(self._expand_key(key)), value)

    def __getitem__(self, key):
        return self._get(_resolve_index(self._expand_key(key)))

    def _set(self, i, value):
        self._store[i] = value

    def _get(self, i):
        try:
            return self._store[i]
        except KeyError:
            pass

        # Could not find element. Try building it.
        value = self._build(i)
        self._store[i] = value
        return value

    @abc.abstractmethod
    def _build(self, i):  # pragma: no cover
        pass


class LazyVector(LazyTensor):
    """A lazy vector."""

    def __init__(self):
        LazyTensor.__init__(self, 1)
        self._rules = []

    def add_rule(self, indices, builder):
        """Add a building rule.

        Note:
            For performance reasons, `indices` must already be resolved!

        Args:
            indices (set): Domain of the rule.
            builder (function): Function that takes in an index and gives back the
                corresponding element.
        """
        self._rules.append((frozenset(indices), builder))

    def _build(self, i):
        i = i[0]  # This will be a one-tuple.
        for indices, builder in self._rules:
            if i in indices:
                return builder(i)
        raise RuntimeError(f'Could not build value for index "{i}".')


class LazyMatrix(LazyTensor):
    """A lazy matrix."""

    def __init__(self):
        LazyTensor.__init__(self, 2)
        self._left_rules = []
        self._right_rules = []
        self._rules = []

    def add_rule(self, indices, builder):
        """Add a building rule.

        Note:
            For performance reasons, `indices` must already be resolved!

        Args:
            indices (set): Domain of the rule.
            builder (function): Function that takes in an index and gives back the
                corresponding element.
        """
        self._rules.append((frozenset(indices), builder))

    def add_left_rule(self, i_left, indices, builder):
        """Add a building rule for a given left index.

        Note:
            For performance reasons, `indices` must already be resolved!

        Args:
            i_left (int): Fixed left index.
            indices (set): Domain of the rule.
            builder (function): Function that takes in a right index and gives back the
                corresponding element.
        """
        self._left_rules.append((i_left, frozenset(indices), builder))

    def add_right_rule(self, i_right, indices, builder):
        """Add a building rule for a given right index.

        Note:
            For performance reasons, `indices` must already be resolved!

        Args:
            i_right (int): Fixed right index.
            indices (set): Domain of the rule.
            builder (function): Function that takes in a left index and gives back the
                corresponding element.
        """
        self._right_rules.append((i_right, frozenset(indices), builder))

    def _build(self, i):
        i_left, i_right = i

        # Check universal rules.
        for indices, builder in self._rules:
            if i_left in indices and i_right in indices:
                return builder(i_left, i_right)

        # Check left rules.
        for i_left_rule, indices, builder in self._left_rules:
            if i_left == i_left_rule and i_right in indices:
                return builder(i_right)

        # Check right rules.
        for i_right_rule, indices, builder in self._right_rules:
            if i_left in indices and i_right == i_right_rule:
                return builder(i_left)

        raise RuntimeError(f"Could not build value for index {i}.")



import logging
from types import FunctionType

from fdm import central_fdm
from lab import B
from matrix import Diagonal, AbstractMatrix, Constant
from plum import Dispatcher, Union, convert

from . import PromisedFDD, PromisedGP
from .input import MultiInput
from .kernel import (
    ZeroKernel,
    PosteriorKernel,
    TensorProductKernel,
    CorrectiveKernel,
    OneKernel,
)
from .lazy import LazyVector, LazyMatrix
from .mean import PosteriorMean, ZeroMean, OneMean
from .mokernel import MultiOutputKernel as MOK
from .momean import MultiOutputMean as MOM
from .random import Random, RandomProcess, Normal
from .util import uprank, num_elements

__all__ = [
    "GP",
    "cross",
    "Measure",
    "AbstractObservations",
    "Observations",
    "Obs",
    "SparseObservations",
    "SparseObs",
]

_dispatch = Dispatcher()

log = logging.getLogger(__name__)


def assert_same_measure(*ps):
    """Assert that processes are associated to the same measure.

    Args:
        *ps (:class:`.measure.GP`): Processes.
    """
    # First check that all processes are constructed from the same measure.
    for p in ps[1:]:
        if ps[0].measure != p.measure:
            raise AssertionError(
                f"Processes {ps[0]} and {p} are associated to different measures."
            )


def intersection_measure_group(*ps):
    """Assert that processes have equal measure groups.

    Args:
        *ps (:class:`.measure.GP`): Processes.
    """
    assert_same_measure(*ps)
    intersection = set(ps[0]._measures)
    for p in ps[1:]:
        intersection &= set(p._measures)
    return intersection


class AbstractObservations:
    """Abstract base class for observations.

    Attributes:
        fdd (:class:`.measure.FDD`): FDD of observations.
        y (column vector): Values of observations.
    """

    @_dispatch
    def __init__(self, fdd: Union[PromisedFDD], y: Union[B.Numeric, AbstractMatrix]):
        self.fdd = fdd
        self.y = y

    @_dispatch
    def __init__(self, *pairs: tuple):
        fdds, ys = zip(*pairs)
        self.fdd = cross(*[fdd.p for fdd in fdds])(MultiInput(*fdds))
        self.y = B.concat(*[uprank(y) for y in ys], axis=0)

    def posterior_kernel(self, measure, p_i, p_j):  # pragma: no cover
        """Get the posterior kernel between two processes.

        Args:
            measure (:class:`.measure.Measure`): Prior.
            p_i (:class:`.measure.GP`): First process.
            p_j (:class:`.measure.GP`): Second process.

        Returns:
            :class:`.kernel.Kernel`: Posterior kernel between the first and
                second process.
        """
        raise NotImplementedError("Posterior kernel construction not implemented.")

    def posterior_mean(self, measure, p):  # pragma: no cover
        """Get the posterior kernel of a process.

        Args:
            measure (:class:`.measure.Measure`): Prior.
            p (:class:`.measure.GP`): Process.

        Returns:
            :class:`.mean.Mean`: Posterior mean of `p`.
        """
        raise NotImplementedError("Posterior mean construction not implemented.")


class Observations(AbstractObservations):
    """Observations.

    Can alternatively construct an instance of `Observations` with tuples of valid
    constructors.

    Args:
        fdd (:class:`.measure.FDD`): FDDs to corresponding to the observations.
        y (tensor): Values of observations.
    """

    def __init__(self, *args):
        AbstractObservations.__init__(self, *args)
        self._K_x_store = {}

    def K_x(self, measure):
        """Kernel matrix of the data.

        Args:
            measure (:class:`.measure.Measure`): Measure.

        Returns:
            matrix: Kernel matrix.
        """
        try:
            return self._K_x_store[id(measure)]
        except KeyError:
            K_x = measure.kernels[self.fdd.p](self.fdd.x)
            self._K_x_store[id(measure)] = K_x
            return K_x

    def posterior_kernel(self, measure, p_i, p_j):
        if num_elements(self.fdd.x) == 0:
            # There are no observations! Just return prior.
            return measure.kernels[p_i, p_j]

        return PosteriorKernel(
            measure.kernels[p_i, p_j],
            measure.kernels[self.fdd.p, p_i],
            measure.kernels[self.fdd.p, p_j],
            self.fdd.x,
            self.K_x(measure),
        )

    def posterior_mean(self, measure, p):
        if num_elements(self.fdd.x) == 0:
            # There are no observations! Just return prior.
            return measure.means[p]

        return PosteriorMean(
            measure.means[p],
            measure.means[self.fdd.p],
            measure.kernels[self.fdd.p, p],
            self.fdd.x,
            self.K_x(measure),
            self.y,
        )


class SparseObservations(AbstractObservations):
    """Observations through inducing points.

    Takes further arguments according to the constructor of
    :class:`.measure.Observations`.

    Args:
        u (:class:`.measure.FDD`): Inducing points
        e (:class:`.measure.GP`): Additive, independent noise process.
    """

    @_dispatch
    def __init__(self, u: Union[tuple, PromisedFDD], *pairs: tuple):
        es, fdds, ys = zip(*pairs)

        # Copy the noises to a measure under which they are independent.
        measure = Measure()
        e = cross(*[GP(e.mean, e.kernel, measure=measure) for e in es])

        fdd = cross(*[fdd.p for fdd in fdds])(MultiInput(*fdds))
        y = B.concat(*[uprank(y) for y in ys], axis=0)
        SparseObservations.__init__(self, u, e, fdd, y)

    @_dispatch
    def __init__(self, us: tuple, e: PromisedGP, fdd: PromisedFDD, y: B.Numeric):
        u = cross(*[u.p for u in us])(MultiInput(*us))
        SparseObservations.__init__(self, u, e, fdd, y)

    @_dispatch
    def __init__(self, u: PromisedFDD, e: PromisedGP, fdd: PromisedFDD, y: B.Numeric):
        AbstractObservations.__init__(self, fdd, y)
        self.u = u
        self.e = e
        self._K_z_store = {}
        self._elbo_store = {}
        self._mu_store = {}
        self._A_store = {}

    def K_z(self, measure):
        """Kernel matrix of the data.

        Args:
            measure (:class:`.measure.Measure`): Measure.

        Returns:
            matrix: Kernel matrix.
        """
        try:
            return self._K_z_store[id(measure)]
        except KeyError:
            self._compute(measure)
            return self._K_z_store[id(measure)]

    def elbo(self, measure):
        """ELBO.

        Args:
            measure (:class:`.measure.Measure`): Measure.

        Returns:
            scalar: ELBO.
        """
        try:
            return self._elbo_store[id(measure)]
        except KeyError:
            self._compute(measure)
            return self._elbo_store[id(measure)]

    def mu(self, measure):
        """Mean of optimal approximating distribution.

        Args:
            measure (:class:`.measure.Measure`): Measure.

        Returns:
            matrix: Mean.
        """
        try:
            return self._mu_store[id(measure)]
        except KeyError:
            self._compute(measure)
            return self._mu_store[id(measure)]

    def A(self, measure):
        """Parameter of the corrective variance of the kernel of the optimal
        approximating distribution.

        Args:
            measure (:class:`.measure.Measure`): Measure.

        Returns:
            matrix: Corrective variance.
        """
        try:
            return self._A_store[id(measure)]
        except KeyError:
            self._compute(measure)
            return self._A_store[id(measure)]

    def _compute(self, measure):
        # Extract processes and inputs.
        p_x, x = self.fdd.p, self.fdd.x
        p_z, z = self.u.p, self.u.x

        # Construct the necessary kernel matrices.
        K_zx = measure.kernels[p_z, p_x](z, x)
        K_z = convert(measure.kernels[p_z](z), AbstractMatrix)
        self._K_z_store[id(measure)] = K_z

        # Evaluating `e.kernel(x)` will yield incorrect results if `x` is a
        # `MultiInput`, because `x` then still designates the particular components
        # of `f`. Fix that by instead designating the elements of `e`.
        if isinstance(x, MultiInput):
            x_n = MultiInput(*(e(fdd.x) for e, fdd in zip(self.e.kernel.ps, x.get())))
        else:
            x_n = x

        # Construct the noise kernel matrix.
        K_n = self.e.kernel(x_n)

        # The approximation can only handle diagonal noise matrices.
        if not isinstance(K_n, Diagonal):
            raise RuntimeError("Kernel matrix of noise must be diagonal.")

        # And construct the components for the inducing point approximation.
        L_z = B.cholesky(K_z)
        A = B.add(B.eye(K_z), B.iqf(K_n, B.transpose(B.solve(L_z, K_zx))))
        self._A_store[id(measure)] = A
        y_bar = uprank(self.y) - self.e.mean(x_n) - measure.means[p_x](x)
        prod_y_bar = B.solve(L_z, B.iqf(K_n, B.transpose(K_zx), y_bar))

        # Compute the optimal mean.
        mu = B.add(
            measure.means[p_z](z),
            B.iqf(A, B.solve(L_z, K_z), prod_y_bar),
        )
        self._mu_store[id(measure)] = mu

        # Compute the ELBO.
        # NOTE: The calculation of `trace_part` asserts that `K_n` is diagonal.
        # The rest, however, is completely generic.
        trace_part = B.ratio(
            Diagonal(measure.kernels[p_x].elwise(x)[:, 0])
            - Diagonal(B.iqf_diag(K_z, K_zx)),
            K_n,
        )
        det_part = B.logdet(2 * B.pi * K_n) + B.logdet(A)
        iqf_part = B.iqf(K_n, y_bar)[0, 0] - B.iqf(A, prod_y_bar)[0, 0]
        self._elbo_store[id(measure)] = -0.5 * (trace_part + det_part + iqf_part)

    def posterior_kernel(self, measure, p_i, p_j):
        return PosteriorKernel(
            measure.kernels[p_i, p_j],
            measure.kernels[self.u.p, p_i],
            measure.kernels[self.u.p, p_j],
            self.u.x,
            self.K_z(measure),
        ) + CorrectiveKernel(
            measure.kernels[self.u.p, p_i],
            measure.kernels[self.u.p, p_j],
            self.u.x,
            self.A(measure),
            self.K_z(measure),
        )

    def posterior_mean(self, measure, p):
        return PosteriorMean(
            measure.means[p],
            measure.means[self.u.p],
            measure.kernels[self.u.p, p],
            self.u.x,
            self.K_z(measure),
            self.mu(measure),
        )


Obs = Observations  #: Shorthand for `Observations`.
SparseObs = SparseObservations  #: Shorthand for `SparseObservations`.


class Measure:
    """A GP model.

    Attributes:
        ps (list[:class:`.measure.GP`]): Processes of the measure.
        mean (:class:`.lazy.LazyVector`): Mean.
        kernels (:class:`.lazy.LazyMatrix`): Kernels.
    """

    def __init__(self):
        self.ps = []
        self._pids = set()
        self.means = LazyVector()
        self.kernels = LazyMatrix()

        # Store named GPs in both ways.
        self._gps_by_name = {}
        self._names_by_gp = {}

    def __hash__(self):
        # This is needed for :func:`.measure.intersection_measure_group`, which puts
        # many :class:`.measure.Measure`s in a `set`. Every measure is unique.
        return id(self)

    @_dispatch
    def __getitem__(self, name: str):
        return self._gps_by_name[name]

    @_dispatch
    def __getitem__(self, p: PromisedGP):
        return self._names_by_gp[id(p)]

    @_dispatch
    def name(self, p: PromisedGP, name: str):
        """Name a GP.

        Args:
            p (:class:`.measure.GP`): GP to name.
            name (str): Name. Must be unique.
        """
        # Delete any existing names and back-references for the GP.
        if id(p) in self._names_by_gp:
            del self._gps_by_name[self._names_by_gp[id(p)]]
            del self._names_by_gp[id(p)]

        # Check that name is not in use.
        if name in self._gps_by_name:
            raise RuntimeError(
                f'Name "{name}" for "{p}" already taken by "{self[name]}".'
            )

        # Set the name and the back-reference.
        self._gps_by_name[name] = p
        self._names_by_gp[id(p)] = name

    def _add_p(self, p):
        # Attach process to measure.
        self.ps.append(p)
        self._pids.add(id(p))
        # Add measure to list of measures of process.
        p._measures.append(self)

    def _update(self, p, mean, kernel, left_rule, right_rule=None):
        # Update means.
        self.means[p] = mean

        # Update kernels.
        self.kernels[p] = kernel
        self.kernels.add_left_rule(id(p), self._pids, left_rule)
        if right_rule:
            self.kernels.add_right_rule(id(p), self._pids, right_rule)
        else:
            self.kernels.add_right_rule(
                id(p), self._pids, lambda i: reversed(self.kernels[p, i])
            )

        # Only now add `p`: `self.pids` above needs to not include `id(p)`.
        self._add_p(p)

        return p

    @_dispatch
    def __call__(self, p: PromisedGP):
        # Make a new GP with `self` as the prior.
        p_copy = GP()
        return self._update(
            p_copy,
            self.means[p],
            self.kernels[p],
            # `p_copy` acts like `p`.
            lambda j: self.kernels[p, j],  # Left rule
            lambda i: self.kernels[i, p],  # Right rule
        )

    @_dispatch
    def __call__(self, fdd: PromisedFDD):
        return Normal(self.means[fdd.p](fdd.x), self.kernels[fdd.p](fdd.x))

    def add_independent_gp(self, p, mean, kernel):
        """Add an independent GP to the model.

        Args:
            p (:class:`.measure.GP`): GP to add.
            mean (:class:`.mean.Mean`): Mean function of GP.
            kernel (:class:`.kernel.Kernel`): Kernel function of GP.

        Returns:
            :class:`.measure.GP`: The newly added independent GP.
        """
        # Update means.
        self.means[p] = mean

        # Update kernels.
        self.kernels[p] = kernel
        self.kernels.add_left_rule(id(p), self._pids, lambda j: ZeroKernel())
        self.kernels.add_right_rule(id(p), self._pids, lambda i: ZeroKernel())

        # Only now add `p`: `self.pids` above needs to not include `id(p)`.
        self._add_p(p)

        return p

    @_dispatch
    def sum(self, p_sum: PromisedGP, other, p: PromisedGP):
        """Sum a GP from the graph with another object.

        Args:
            p_sum (:class:`.measure.GP`): GP that is the sum.
            obj1 (other type or :class:`.measure.GP`): First term in the sum.
            obj2 (other type or :class:`.measure.GP`): Second term in the sum.

        Returns:
            :class:`.measure.GP`: The GP corresponding to the sum.
        """
        return self.sum(p_sum, p, other)

    @_dispatch
    def sum(
        self, p_sum: PromisedGP, p: PromisedGP, other: Union[B.Numeric, FunctionType]
    ):
        return self._update(
            p_sum,
            self.means[p] + other,
            self.kernels[p],
            lambda j: self.kernels[p, j],
        )

    @_dispatch
    def sum(self, p_sum: PromisedGP, p1: PromisedGP, p2: PromisedGP):
        assert_same_measure(p1, p2)
        return self._update(
            p_sum,
            self.means[p1] + self.means[p2],
            (
                self.kernels[p1]
                + self.kernels[p2]
                + self.kernels[p1, p2]
                + self.kernels[p2, p1]
            ),
            lambda j: self.kernels[p1, j] + self.kernels[p2, j],
        )

    @_dispatch
    def mul(self, p_mul: PromisedGP, other, p: PromisedGP):
        """Multiply a GP from the graph with another object.

        Args:
            p_mul (:class:`.measure.GP`): GP that is the product.
            obj1 (object): First factor in the product.
            obj2 (object): Second factor in the product.
            other (object): Other object in the product.

        Returns:
            :class:`.measure.GP`: The GP corresponding to the product.
        """
        return self.mul(p_mul, p, other)

    @_dispatch
    def mul(self, p_mul: PromisedGP, p: PromisedGP, other: B.Numeric):
        return self._update(
            p_mul,
            self.means[p] * other,
            self.kernels[p] * other ** 2,
            lambda j: self.kernels[p, j] * other,
        )

    @_dispatch
    def mul(self, p_mul: PromisedGP, p: PromisedGP, f: FunctionType):
        def ones(x):
            return Constant(B.one(x), num_elements(x), 1)

        return self._update(
            p_mul,
            f * self.means[p],
            f * self.kernels[p],
            lambda j: TensorProductKernel(f, ones) * self.kernels[p, j],
        )

    @_dispatch
    def mul(self, p_mul: PromisedGP, p1: PromisedGP, p2: PromisedGP):
        assert_same_measure(p1, p2)
        term1 = self.sum(
            GP(),
            self.mul(GP(), lambda x: self.means[p1](x), p2),
            self.mul(GP(), p1, lambda x: self.means[p2](x)),
        )
        term2 = self.add_independent_gp(
            GP(),
            -self.means[p1] * self.means[p2],
            (
                self.kernels[p1] * self.kernels[p2]
                + self.kernels[p1, p2] * self.kernels[p2, p1]
            ),
        )
        return self.sum(p_mul, term1, term2)

    def shift(self, p_shifted, p, shift):
        """Shift a GP.

        Args:
            p_shifted (:class:`.measure.GP`): Shifted GP.
            p (:class:`.measure.GP`): GP to shift.
            shift (object): Amount to shift by.

        Returns:
            :class:`.measure.GP`: The shifted GP.
        """
        return self._update(
            p_shifted,
            self.means[p].shift(shift),
            self.kernels[p].shift(shift),
            lambda j: self.kernels[p, j].shift(shift, 0),
        )

    def stretch(self, p_stretched, p, stretch):
        """Stretch a GP.

        Args:
            p_stretched (:class:`.measure.GP`): Stretched GP.
            p (:class:`.measure.GP`): GP to stretch.
            stretch (object): Extent of stretch.

        Returns:
            :class:`.measure.GP`: The stretched GP.
        """
        return self._update(
            p_stretched,
            self.means[p].stretch(stretch),
            self.kernels[p].stretch(stretch),
            lambda j: self.kernels[p, j].stretch(stretch, 1),
        )

    def select(self, p_selected, p, *dims):
        """Select input dimensions.

        Args:
            p_selected (:class:`.measure.GP`): GP with selected inputs.
            p (:class:`.measure.GP`): GP to select input dimensions from.
            *dims (object): Dimensions to select.

        Returns:
            :class:`.measure.GP`: GP with the specific input dimensions.
        """
        return self._update(
            p_selected,
            self.means[p].select(dims),
            self.kernels[p].select(dims),
            lambda j: self.kernels[p, j].select(dims, None),
        )

    def transform(self, p_transformed, p, f):
        """Transform the inputs of a GP.

        Args:
            p_transformed (:class:`.measure.GP`): GP with transformed inputs.
            p (:class:`.measure.GP`): GP to input transform.
            f (function): Input transform.

        Returns:
            :class:`.measure.GP`: Input-transformed GP.
        """
        return self._update(
            p_transformed,
            self.means[p].transform(f),
            self.kernels[p].transform(f),
            lambda j: self.kernels[p, j].transform(f, None),
        )

    def diff(self, p_diff, p, dim=0):
        """Differentiate a GP.

        Args:
            p_diff (:class:`.measure.GP`): Derivative.
            p (:class:`.measure.GP`): GP to differentiate.
            dim (int, optional): Dimension of feature which to take the derivative
                with respect to. Defaults to `0`.

        Returns:
            :class:`.measure.GP`: Derivative of GP.
        """
        return self._update(
            p_diff,
            self.means[p].diff(dim),
            self.kernels[p].diff(dim),
            lambda j: self.kernels[p, j].diff(dim, None),
        )

    @_dispatch
    def condition(self, obs: AbstractObservations):
        """Condition the measure on observations.

        Args:
            obs (:class:`.measure.AbstractObservations`): Observations to condition on.

        Returns:
            list[:class:`.measure.GP`]: Posterior processes.
        """
        posterior = Measure()
        posterior.ps = list(self.ps)
        posterior._pids = set(self._pids)

        posterior.means.add_rule(posterior._pids, lambda i: obs.posterior_mean(self, i))
        posterior.kernels.add_rule(
            posterior._pids, lambda i, j: obs.posterior_kernel(self, i, j)
        )

        # Update backreferences.
        for p in posterior.ps:
            p._measures.append(posterior)

        return posterior

    @_dispatch
    def condition(self, fdd: PromisedFDD, y: B.Numeric):
        return self.condition(Obs(fdd, y))

    @_dispatch
    def condition(self, pair: tuple):
        return self.condition(Obs(*pair))

    @_dispatch
    def condition(self, *pairs: tuple):
        return self.condition(Obs(*pairs))

    @_dispatch
    def __or__(self, *args):
        return self.condition(*args)

    @_dispatch
    def cross(self, p_cross: PromisedGP, *ps: PromisedGP):
        """Construct the Cartesian product of a collection of processes.

        Args:
            p_cross (:class:`.measure.GP`): GP that is the Cartesian product.
            *ps (:class:`.measure.GP`): Processes to construct the Cartesian product of.

        Returns:
            :class:`.measure.GP`: The Cartesian product of `ps`.
        """
        mok = MOK(self, *ps)
        return self._update(
            p_cross,
            MOM(self, *ps),
            mok,
            lambda j: mok.transform(None, lambda y: FDD(j, y)),
        )

    @_dispatch
    def sample(self, n: int, *fdds: PromisedFDD):
        """Sample multiple processes simultaneously.

        Args:
            n (int, optional): Number of samples. Defaults to `1`.
            *fdds (:class:`.measure.FDD`): Locations to sample at.

        Returns:
            tuple: Tuple of samples.
        """
        sample = cross(*self.ps)(MultiInput(*fdds)).sample(n)

        # Unpack sample.
        lengths = [num_elements(fdd) for fdd in fdds]
        i, samples = 0, []
        for length in lengths:
            samples.append(sample[i : i + length, :])
            i += length
        return samples[0] if len(samples) == 1 else samples

    @_dispatch
    def sample(self, *fdds: PromisedFDD):
        return self.sample(1, *fdds)

    @_dispatch
    def logpdf(self, *pairs: Union[list, tuple]):
        """Compute the logpdf of one multiple observations.

        Can also give an `AbstractObservations`.

        Args:
            *pairs (tuple[:class:`.measure.FDD`, tensor]): Pairs of FDDs and values
                of the observations.

        Returns:
            scalar: Logpdf.
        """
        fdds, ys = zip(*pairs)
        y = B.concat(*[uprank(y) for y in ys], axis=0)
        return self(
            self.cross(GP(), *[fdd.p for fdd in fdds])(MultiInput(*fdds))
        ).logpdf(y)

    @_dispatch
    def logpdf(self, fdd: PromisedFDD, y: B.Numeric):
        return self(fdd).logpdf(y)

    @_dispatch
    def logpdf(self, obs: Observations):
        return self.logpdf(obs.fdd, obs.y)

    @_dispatch
    def logpdf(self, obs: SparseObservations):
        return obs.elbo(self)


class GP(RandomProcess):
    """Gaussian process.

    Args:
        mean (:class:`.mean.Mean`, optional): Mean function of the process. Defaults
            to zero.
        kernel (:class:`.kernel.Kernel`): Kernel of the process.
        measure (:class:`.measure.Measure`): Measure to attach to. Must be given as
            a keyword argument.
        name (:obj:`str`, optional): Name. Must be given as a keyword argment.
    """

    @_dispatch
    def __init__(self, mean, kernel, measure=None, name=None):
        self._measures = []

        # If no measure is given, create one.
        if measure is None:
            measure = Measure()

        # Resolve mean.
        if isinstance(mean, (B.Numeric, FunctionType)):
            mean = mean * OneMean()

        # Resolve kernel.
        if isinstance(kernel, (B.Numeric, FunctionType)):
            kernel = kernel * OneKernel()

        # Add a new `GP` to the measure with the resolved kernel and mean. The measure
        # will add itself to `self.measures`.
        measure.add_independent_gp(self, mean, kernel)

        # If a name is given, set the name.
        if name:
            measure.name(self, name)

    @_dispatch
    def __init__(self, kernel, measure=None, name=None):
        self.__init__(ZeroMean(), kernel, measure=measure, name=name)

    @_dispatch
    def __init__(self):
        self._measures = []

    @property
    def measure(self):
        """Measure that the GP was constructed with."""
        if len(self._measures) == 0:
            raise RuntimeError("GP is not associated to a measure.")
        else:
            return self._measures[0]

    @property
    def kernel(self):
        """Kernel of the GP."""
        return self.measure.kernels[self]

    @property
    def mean(self):
        """Mean function of the GP."""
        return self.measure.means[self]

    @property
    def name(self):
        """Name of the GP."""
        return self.measure[self]

    @name.setter
    @_dispatch
    def name(self, name: str):
        for measure in self._measures:
            measure.name(self, name)

    def __call__(self, x):
        """Construct a finite-dimensional distribution at specified locations.

        Args:
            x (input): Points to construct the distribution at.

        Returns:
            :class:`.measure.FDD`: Finite-dimensional distribution.
        """
        return FDD(self, x)

    @_dispatch
    def __add__(self, other: Union[B.Numeric, FunctionType]):
        res = GP()
        for measure in self._measures:
            measure.sum(res, self, other)
        return res

    @_dispatch
    def __add__(self, other: "GP"):
        res = GP()
        for measure in intersection_measure_group(self, other):
            measure.sum(res, self, other)
        return res

    @_dispatch
    def __mul__(self, other: Union[B.Numeric, FunctionType]):
        res = GP()
        for measure in self._measures:
            measure.mul(res, self, other)
        return res

    @_dispatch
    def __mul__(self, other: "GP"):
        res = GP()
        for measure in intersection_measure_group(self, other):
            measure.mul(res, self, other)
        return res

    def shift(self, shift):
        """Shift the GP. See :meth:`.measure.Graph.shift`."""
        res = GP()
        for measure in self._measures:
            measure.shift(res, self, shift)
        return res

    def stretch(self, stretch):
        """Stretch the GP. See :meth:`.measure.Graph.stretch`."""
        res = GP()
        for measure in self._measures:
            measure.stretch(res, self, stretch)
        return res

    def __gt__(self, stretch):
        """Shorthand for :meth:`.measure.GP.stretch`."""
        return self.stretch(stretch)

    def transform(self, f):
        """Input transform the GP. See :meth:`.measure.Graph.transform`."""
        res = GP()
        for measure in self._measures:
            measure.transform(res, self, f)
        return res

    def select(self, *dims):
        """Select dimensions from the input. See :meth:`.measure.Graph.select`."""
        res = GP()
        for measure in self._measures:
            measure.select(res, self, *dims)
        return res

    def __getitem__(self, *dims):
        """Shorthand for :meth:`.measure.GP.select`."""
        return self.select(*dims)

    def diff(self, dim=0):
        """Differentiate the GP. See :meth:`.measure.Graph.diff`."""
        res = GP()
        for measure in self._measures:
            measure.diff(res, self, dim)
        return res

    def diff_approx(self, deriv=1, order=6):
        """Approximate the derivative of the GP by constructing a finite
        difference approximation.

        Args:
            deriv (int, optional): Order of the derivative. Defaults to `1`.
            order (int, optional): Order of the estimate. Defaults to `6`.

        Returns:
            :class:`.measure.GP`: Approximation of the derivative of the GP.
        """
        # Use the FDM library to figure out the coefficients.
        fdm = central_fdm(order, deriv, adapt=0, factor=1e8)
        fdm.estimate()  # Estimate step size.

        # Construct finite difference.
        df = 0
        for g, c in zip(fdm.grid, fdm.coefs):
            df += c * self.shift(-g * fdm.step)
        return df / fdm.step ** deriv

    @property
    def stationary(self):
        """Stationarity of the GP."""
        return self.kernel.stationary

    def __str__(self):
        return self.display()

    def __repr__(self):
        return self.display()

    def display(self, formatter=lambda x: x):
        """Display the GP.

        Args:
            formatter (function, optional): Function to format values.

        Returns:
            str: GP as a string.
        """
        if self._measures:
            return (
                f"GP({self.mean.display(formatter)}, {self.kernel.display(formatter)})"
            )
        else:
            return "GP()"


PromisedGP.deliver(GP)


def cross(*ps):
    """Construct the Cartesian product of a collection of processes.

    Args:
        *ps (:class:`.measure.GP`): Processes to construct the Cartesian product of.

    Returns:
        :class:`.measure.GP`: The Cartesian product of `ps`.
    """
    p_cross = GP()
    for measure in intersection_measure_group(*ps):
        measure.cross(p_cross, *ps)
    return p_cross


class FDD(Normal):
    """Finite-dimensional distribution.

    Args:
        p (:class:`.random.RandomProcess`): Process of FDD.
        x (input): Inputs that `p` is evaluated at.

    Attributes:
        p (:class:`.random.RandomProcess`): Process of FDD.
        x (input): Inputs that `p` is evaluated at.
    """

    def __init__(self, p, x):
        self.p = p
        self.x = x
        Normal.__init__(self, lambda: p.mean(x), lambda: p.kernel(x))

    def __str__(self):
        return f"FDD({self.p}, {self.x})"

    def __repr__(self):
        return f"FDD({self.p!r}, {self.x!r})"


PromisedFDD.deliver(FDD)


@num_elements.dispatch
def num_elements(fdd: FDD):
    return num_elements(fdd.x)



import logging

import alge
import numpy as np
from algebra.util import identical, to_tensor
from lab import B
from matrix import Dense, LowRank, Diagonal, Zero, Constant, AbstractMatrix
from plum import Dispatcher, Self, convert, Union

from . import PromisedFDD as FDD
from .input import Input, Unique, WeightedUnique, MultiInput
from .util import num_elements, uprank

__all__ = [
    "Kernel",
    "OneKernel",
    "ZeroKernel",
    "ScaledKernel",
    "EQ",
    "RQ",
    "Matern12",
    "Exp",
    "Matern32",
    "Matern52",
    "Delta",
    "FixedDelta",
    "Linear",
    "DerivativeKernel",
    "DecayingKernel",
    "LogKernel",
]

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


class Delta(Kernel):
    """Kronecker delta kernel.

    Args:
        epsilon (float, optional): Tolerance for equality in squared distance.
            Defaults to `1e-10`.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, epsilon=1e-10):
        self.epsilon = epsilon

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y:
            return B.fill_diag(B.one(x), num_elements(x))
        else:
            return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch(Unique, Unique)
    def __call__(self, x, y):
        x, y = x.get(), y.get()
        if x is y:
            return B.fill_diag(B.one(x), num_elements(x))
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch(WeightedUnique, WeightedUnique)
    def __call__(self, x, y):
        w_x, w_y = x.w, y.w
        x, y = x.get(), y.get()
        if x is y:
            return Diagonal(1 / w_x)
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch(Unique, object)
    def __call__(self, x, y):
        x = x.get()
        return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch(object, Unique)
    def __call__(self, x, y):
        y = y.get()
        return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch(Unique, Unique)
    def elwise(self, x, y):
        x, y = x.get(), y.get()
        if x is y:
            return B.ones(B.dtype(x), num_elements(x), 1)
        else:
            return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch(WeightedUnique, WeightedUnique)
    def elwise(self, x, y):
        w_x, w_y = x.w, y.w
        x, y = x.get(), y.get()
        if x is y:
            return uprank(1 / w_x)
        else:
            return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch(Unique, object)
    def elwise(self, x, y):
        x = x.get()
        return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch(object, Unique)
    def elwise(self, x, y):
        return B.zeros(B.dtype(x), num_elements(x), 1)

    @_dispatch(B.Numeric, B.Numeric)
    def elwise(self, x, y):
        if x is y:
            return B.ones(B.dtype(x), num_elements(x), 1)
        else:
            return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        dtype = B.dtype(dists2)
        return B.cast(dtype, B.lt(dists2, B.cast(dtype, self.epsilon)))

    @property
    def _stationary(self):
        return True

    @_dispatch(Self)
    def __eq__(self, other):
        return self.epsilon == other.epsilon


class FixedDelta(Kernel):
    """Kronecker delta kernel that produces a diagonal matrix with given
    noises if and only if the inputs are identical and of the right shape.

    Args:
        noises (vector): Noises.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, noises):
        self.noises = noises

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y and num_elements(x) == B.shape(self.noises)[0]:
            return Diagonal(self.noises)
        else:
            return Zero(B.dtype(x), num_elements(x), num_elements(y))

    @_dispatch(B.Numeric, B.Numeric)
    def elwise(self, x, y):
        if x is y and num_elements(x) == B.shape(self.noises)[0]:
            return uprank(self.noises)
        else:
            return Zero(B.dtype(x), num_elements(x), 1)

    @property
    def _stationary(self):
        return True

    @_dispatch(Self)
    def __eq__(self, other):
        return B.all(self.noises == other.noises)

class PosteriorKernel(Kernel):
    """Posterior kernel.

    Args:
        k_ij (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the left input and the right input respectively.
        k_zi (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the left input respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the right input respectively.
        z (input): Locations of data.
        K_z (matrix): Kernel matrix of data.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k_ij, k_zi, k_zj, z, K_z):
        self.k_ij = k_ij
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.K_z = convert(K_z, AbstractMatrix)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.subtract(
            self.k_ij(x, y), B.iqf(self.K_z, self.k_zi(self.z, x), self.k_zj(self.z, y))
        )

    @_dispatch(object, object)
    def elwise(self, x, y):
        iqf_diag = B.iqf_diag(self.K_z, self.k_zi(self.z, x), self.k_zj(self.z, y))
        return B.subtract(self.k_ij.elwise(x, y), B.expand_dims(iqf_diag, axis=1))


class CorrectiveKernel(Kernel):
    """Kernel that adds the corrective variance in sparse conditioning.

    Args:
        k_zi (:class:`.kernel.Kernel`): Kernel between the processes corresponding to
            the left input and the inducing points respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between the processes corresponding to
            the right input and the inducing points respectively.
        z (input): Locations of the inducing points.
        A (tensor): Corrective matrix.
        L (tensor): Kernel matrix of the inducing points.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k_zi, k_zj, z, A, K_z):
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.A = A
        self.L = B.cholesky(convert(K_z, AbstractMatrix))

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.iqf(
            self.A,
            B.solve(self.L, self.k_zi(self.z, x)),
            B.solve(self.L, self.k_zj(self.z, y)),
        )

    @_dispatch(object, object)
    def elwise(self, x, y):
        return B.iqf_diag(
            self.A,
            B.solve(self.L, self.k_zi(self.z, x)),
            B.solve(self.L, self.k_zj(self.z, y)),
        )[:, None]




# noinspection PyUnresolvedReferences
import matrix
# noinspection PyUnresolvedReferences
from lab import B

from plum import PromisedType

PromisedFDD = PromisedType()
PromisedGP = PromisedType()

from .input import *
from .kernel import *
from .lazy import *
from .mean import *
from .measure import *
from .mokernel import *
from .momean import *
from .random import *
from .util import *




from lab import B

from stheno.util import uprank, num_elements
from .util import approx


def test_num_elements():
    assert num_elements(B.randn()) == 1
    assert num_elements(B.randn(2)) == 2
    assert num_elements(B.randn(2, 2)) == 2


def test_uprank():
    # Check decorative behaviour.
    @uprank
    def with_uprank(x, y):
        return x, y

    # Check upranking behaviour.
    approx(with_uprank(0, B.zeros(1)), ([[0]], [[0]]))
    approx(with_uprank(B.zeros(1), B.zeros(1)), ([[0]], [[0]]))
    approx(with_uprank(B.zeros(1, 1), B.zeros(1)), ([[0]], [[0]]))

    # Check that it leaves objects alone.
    x = []
    assert uprank(x) is x



import pytest
import lab as B

from stheno.measure import Measure, GP
from stheno.input import Input, MultiInput
from stheno.kernel import EQ
from stheno.momean import MultiOutputMean
from .util import approx


@pytest.mark.parametrize("x", [B.linspace(0, 1, 10), Input(B.linspace(0, 1, 10))])
def test_momean(x):
    prior = Measure()
    p1 = GP(lambda x: 2 * x, 1 * EQ(), measure=prior)
    p2 = GP(1, 2 * EQ().stretch(2), measure=prior)

    m = MultiOutputMean(prior, p1, p2)
    ms = prior.means

    # Check representation.
    assert str(m) == "MultiOutputMean(<lambda>, 1)"

    # Check computation.
    approx(m(x), B.concat(ms[p1](x), ms[p2](x), axis=0))
    approx(m(p1(x)), ms[p1](x))
    approx(m(p2(x)), ms[p2](x))
    approx(m(MultiInput(p2(x), p1(x))), B.concat(ms[p2](x), ms[p1](x), axis=0))



import numpy as np
import pytest
import tensorflow as tf
from lab.tensorflow import B
from matrix import Diagonal, Zero

from stheno.input import Input, Unique, WeightedUnique, MultiInput
from stheno.kernel import (
    EQ,
    RQ,
    Matern12,
    Matern32,
    Matern52,
    Delta,
    FixedDelta,
    Kernel,
    Linear,
    OneKernel,
    ZeroKernel,
    PosteriorKernel,
    ShiftedKernel,
    TensorProductKernel,
    CorrectiveKernel,
    DecayingKernel,
    LogKernel,
    perturb,
    num_elements,
)
from stheno.measure import FDD
from .util import approx


def standard_kernel_tests(k, shapes=None, dtype=np.float64):
    if shapes is None:
        shapes = [
            ((10, 2), (5, 2)),
            ((10, 1), (5, 1)),
            ((10,), (5, 1)),
            ((10, 1), (5,)),
            ((10,), (5,)),
            ((10,), ()),
            ((), (5,)),
            ((), ()),
        ]

    # Check various shapes of arguments.
    for shape1, shape2 in shapes:
        x1 = B.randn(dtype, *shape1)
        x2 = B.randn(dtype, *shape2)

        # Check that the kernel computes consistently.
        approx(k(x1, x2), B.transpose(reversed(k)(x2, x1)))

        # Check `elwise`.
        x2 = B.randn(dtype, *shape1)
        approx(k.elwise(x1, x2), B.diag(k(x1, x2))[:, None])
        # Check againtst fallback brute force computation.
        approx(k.elwise(x1, x2), Kernel.elwise(k, x1, x2))
        # The element-wise computation is more accurate, which is why we allow a
        # discrepancy a bit larger than the square root of the machine epsilon.
        approx(k.elwise(x1)[:, 0], B.diag(k(x1)), atol=1e-6, rtol=1e-6)
        approx(k.elwise(x1), Kernel.elwise(k, x1))


def test_corner_cases():
    with pytest.raises(RuntimeError):
        Kernel()(1.0)


@pytest.mark.parametrize(
    "x1", [B.randn(10), Input(B.randn(10)), FDD(None, B.randn(10))]
)
@pytest.mark.parametrize(
    "x2", [B.randn(10), Input(B.randn(10)), FDD(None, B.randn(10))]
)
def test_construction(x1, x2):
    k = EQ()

    k(x1)
    k(x1, x2)

    k.elwise(x1)
    k.elwise(x1, x2)

    # Test `MultiInput` construction.
    approx(
        k(MultiInput(x1, x2)),
        B.concat2d([k(x1, x1), k(x1, x2)], [k(x2, x1), k(x2, x2)]),
    )
    approx(k(x1, MultiInput(x1, x2)), B.concat(k(x1, x1), k(x1, x2), axis=1))
    approx(k(MultiInput(x1, x2), x2), B.concat(k(x1, x2), k(x2, x2), axis=0))

    approx(
        k.elwise(MultiInput(x1, x2)),
        B.concat(k.elwise(x1, x1), k.elwise(x2, x2), axis=0),
    )
    with pytest.raises(ValueError):
        k.elwise(MultiInput(x1), MultiInput(x1, x2))
    with pytest.raises(ValueError):
        k.elwise(x1, MultiInput(x1, x2))
    with pytest.raises(ValueError):
        k.elwise(MultiInput(x1, x2), x2)


def test_basic_arithmetic():
    k1 = EQ()
    k2 = RQ(1e-1)
    k3 = Matern12()
    k4 = Matern32()
    k5 = Matern52()
    k6 = Delta()
    k7 = Linear()
    xs1 = B.randn(10, 2), B.randn(20, 2)
    xs2 = B.randn(), B.randn()

    approx(k6(xs1[0]), k6(xs1[0], xs1[0]))
    approx((k1 * k2)(*xs1), k1(*xs1) * k2(*xs1))
    approx((k1 * k2)(*xs2), k1(*xs2) * k2(*xs2))
    approx((k3 + k4)(*xs1), k3(*xs1) + k4(*xs1))
    approx((k3 + k4)(*xs2), k3(*xs2) + k4(*xs2))
    approx((5.0 * k5)(*xs1), 5.0 * k5(*xs1))
    approx((5.0 * k5)(*xs2), 5.0 * k5(*xs2))
    approx((5.0 + k7)(*xs1), 5.0 + k7(*xs1))
    approx((5.0 + k7)(*xs2), 5.0 + k7(*xs2))
    approx(k1.stretch(2.0)(*xs1), k1(xs1[0] / 2.0, xs1[1] / 2.0))
    approx(k1.stretch(2.0)(*xs2), k1(xs2[0] / 2.0, xs2[1] / 2.0))
    approx(k1.periodic(1.0)(*xs1), k1.periodic(1.0)(xs1[0], xs1[1] + 5.0))
    approx(k1.periodic(1.0)(*xs2), k1.periodic(1.0)(xs2[0], xs2[1] + 5.0))


def test_reversal():
    x1 = B.randn(10, 2)
    x2 = B.randn(5, 2)
    x3 = B.randn()

    # Test with a stationary and non-stationary kernel.
    for k in [EQ(), Linear()]:
        approx(k(x1), reversed(k)(x1))
        approx(k(x3), reversed(k)(x3))
        approx(k(x1, x2), reversed(k)(x1, x2))
        approx(k(x1, x2), reversed(k)(x2, x1).T)

        # Test double reversal does the right thing.
        approx(k(x1), reversed(reversed(k))(x1))
        approx(k(x3), reversed(reversed(k))(x3))
        approx(k(x1, x2), reversed(reversed(k))(x1, x2))
        approx(k(x1, x2), reversed(reversed(k))(x2, x1).T)

    # Verify that the kernel has the right properties.
    k = reversed(EQ())
    assert k.stationary

    k = reversed(Linear())
    assert not k.stationary
    assert str(k) == "Reversed(Linear())"

    # Check equality.
    assert reversed(Linear()) == reversed(Linear())
    assert reversed(Linear()) != Linear()
    assert reversed(Linear()) != reversed(EQ())
    assert reversed(Linear()) != reversed(DecayingKernel(1, 1))

    # Standard tests:
    standard_kernel_tests(k)


def test_delta_properties():
    k = Delta()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "Delta()"

    # Check equality.
    assert Delta() == Delta()
    assert Delta() != Delta(epsilon=k.epsilon * 10)
    assert Delta() != EQ()


@pytest.mark.parametrize(
    "x1, w1, x2, w2",
    [
        (np.array(0), np.ones(1), np.array(1), np.ones(1)),
        (B.randn(10), np.ones(10), B.randn(5), np.ones(5)),
        (B.randn(10, 1), np.ones(10), B.randn(5, 1), np.ones(5)),
        (B.randn(10, 2), np.ones(10), B.randn(5, 2), np.ones(5)),
    ],
)
def test_delta_evaluations(x1, w1, x2, w2):
    k = Delta()
    n1 = num_elements(x1)
    n2 = num_elements(x2)

    # Check uniqueness checks.
    approx(k(x1), B.eye(n1))
    approx(k(x1, x2), B.zeros(n1, n2))

    # Standard tests:
    standard_kernel_tests(k)

    # Test `Unique` inputs.
    assert isinstance(k(Unique(x1), Unique(x1.copy())), Zero)
    assert isinstance(k(Unique(x1), Unique(x1)), Diagonal)
    assert isinstance(k(Unique(x1), x1), Zero)
    assert isinstance(k(x1, Unique(x1)), Zero)

    approx(k.elwise(Unique(x1), Unique(x1.copy())), B.zeros(n1, 1))
    approx(k.elwise(Unique(x1), Unique(x1)), B.ones(n1, 1))
    approx(k.elwise(Unique(x1), x1), B.zeros(n1, 1))

    # Test `WeightedUnique` inputs.
    assert isinstance(k(WeightedUnique(x1, w1), WeightedUnique(x1.copy(), w1)), Zero)
    assert isinstance(k(WeightedUnique(x1, w1), WeightedUnique(x1, w1)), Diagonal)
    assert isinstance(k(WeightedUnique(x1, w1), x1), Zero)
    assert isinstance(k(x1, WeightedUnique(x1, w1)), Zero)

    approx(
        k.elwise(WeightedUnique(x1, w1), WeightedUnique(x1.copy(), w1)), B.zeros(n1, 1)
    )
    approx(k.elwise(WeightedUnique(x1, w1), WeightedUnique(x1, w1)), B.ones(n1, 1))
    approx(k.elwise(WeightedUnique(x1, w1), x1), B.zeros(n1, 1))
    approx(k.elwise(x1, WeightedUnique(x1, w1)), B.zeros(n1, 1))
    approx(k.elwise(x1, WeightedUnique(x1, w1)), B.zeros(n1, 1))


def test_fixed_delta():
    noises = B.rand(3)
    k = FixedDelta(noises)

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "FixedDelta()"

    # Check equality.
    assert FixedDelta(noises) == FixedDelta(noises)
    assert FixedDelta(noises) != FixedDelta(2 * noises)
    assert FixedDelta(noises) != EQ()

    # Standard tests:
    standard_kernel_tests(k)

    # Check correctness.
    x1 = B.randn(5)
    x2 = B.randn(5)
    approx(k(x1), B.zeros(5, 5))
    approx(k.elwise(x1), B.zeros(5, 1))
    approx(k(x1, x2), B.zeros(5, 5))
    approx(k.elwise(x1, x2), B.zeros(5, 1))

    x1 = B.randn(3)
    x2 = B.randn(3)
    approx(k(x1), B.diag(noises))
    approx(k.elwise(x1), B.uprank(noises))
    approx(k(x1, x2), B.zeros(3, 3))
    approx(k.elwise(x1, x2), B.zeros(3, 1))


def test_eq():
    k = EQ()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "EQ()"

    # Test equality.
    assert EQ() == EQ()
    assert EQ() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_rq():
    k = RQ(1e-1)

    # Verify that the kernel has the right properties.
    assert k.alpha == 1e-1
    assert k.stationary
    assert str(k) == "RQ(0.1)"

    # Test equality.
    assert RQ(1e-1) == RQ(1e-1)
    assert RQ(1e-1) != RQ(2e-1)
    assert RQ(1e-1) != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_exp():
    k = Matern12()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "Exp()"

    # Test equality.
    assert Matern12() == Matern12()
    assert Matern12() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_mat32():
    k = Matern32()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "Matern32()"

    # Test equality.
    assert Matern32() == Matern32()
    assert Matern32() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_mat52():
    k = Matern52()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "Matern52()"

    # Test equality.
    assert Matern52() == Matern52()
    assert Matern52() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_one():
    k = OneKernel()

    x1 = B.randn(10, 2)
    x2 = B.randn(5, 2)

    # Test that the kernel computes correctly.
    approx(k(x1, x2), np.ones((10, 5)))

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "1"

    # Test equality.
    assert OneKernel() == OneKernel()
    assert OneKernel() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_zero():
    k = ZeroKernel()
    x1 = B.randn(10, 2)
    x2 = B.randn(5, 2)

    # Test that the kernel computes correctly.
    approx(k(x1, x2), np.zeros((10, 5)))

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "0"

    # Test equality.
    assert ZeroKernel() == ZeroKernel()
    assert ZeroKernel() != Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_linear():
    k = Linear()

    # Verify that the kernel has the right properties.
    assert not k.stationary
    assert str(k) == "Linear()"

    # Test equality.
    assert Linear() == Linear()
    assert Linear() != EQ()

    # Standard tests:
    standard_kernel_tests(k)


def test_decaying_kernel():
    k = DecayingKernel(3.0, 4.0)

    # Verify that the kernel has the right properties.
    assert not k.stationary
    assert str(k) == "DecayingKernel(3.0, 4.0)"

    # Test equality.
    assert DecayingKernel(3.0, 4.0) == DecayingKernel(3.0, 4.0)
    assert DecayingKernel(3.0, 4.0) != DecayingKernel(3.0, 5.0)
    assert DecayingKernel(3.0, 4.0) != DecayingKernel(4.0, 4.0)
    assert DecayingKernel(3.0, 4.0) != EQ()

    # Standard tests:
    standard_kernel_tests(k)


def test_log_kernel():
    k = LogKernel()

    # Verify that the kernel has the right properties.
    assert k.stationary
    assert str(k) == "LogKernel()"

    # Test equality.
    assert LogKernel() == LogKernel()
    assert LogKernel() != EQ()

    # Standard tests:
    standard_kernel_tests(k)


def test_posterior_kernel():
    k = PosteriorKernel(EQ(), EQ(), EQ(), B.randn(5, 2), EQ()(B.randn(5, 1)))

    # Verify that the kernel has the right properties.
    assert not k.stationary
    assert str(k) == "PosteriorKernel()"

    # Standard tests:
    standard_kernel_tests(k, shapes=[((10, 2), (5, 2))])


def test_corrective_kernel():
    a, b = B.randn(3, 3), B.randn(3, 3)
    a, b = a.dot(a.T), b.dot(b.T)
    z = B.randn(3, 2)
    k = CorrectiveKernel(EQ(), EQ(), z, a, b)

    # Verify that the kernel has the right properties.
    assert not k.stationary
    assert str(k) == "CorrectiveKernel()"

    # Standard tests:
    standard_kernel_tests(k, shapes=[((10, 2), (5, 2))])


def test_sum():
    k1 = EQ().stretch(2)
    k2 = 3 * RQ(1e-2).stretch(5)
    k = k1 + k2

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert EQ() + Linear() == EQ() + Linear()
    assert EQ() + Linear() == Linear() + EQ()
    assert EQ() + Linear() != EQ() + RQ(1e-1)
    assert EQ() + Linear() != RQ(1e-1) + Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_product():
    k = (2 * EQ().stretch(10)) * (3 * RQ(1e-2).stretch(20))

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert EQ() * Linear() == EQ() * Linear()
    assert EQ() * Linear() == Linear() * EQ()
    assert EQ() * Linear() != EQ() * RQ(1e-1)
    assert EQ() * Linear() != RQ(1e-1) * Linear()

    # Standard tests:
    standard_kernel_tests(k)


def test_stretched():
    k = EQ().stretch(2)

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert EQ().stretch(2) == EQ().stretch(2)
    assert EQ().stretch(2) != EQ().stretch(3)
    assert EQ().stretch(2) != Matern12().stretch(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = EQ().stretch(1, 2)

    # Verify that the kernel has the right properties.
    assert not k.stationary

    # Check passing in a list.
    k = EQ().stretch(np.array([1, 2]))
    k(B.randn(10, 2))


def test_periodic():
    k = EQ().stretch(2).periodic(3)

    # Verify that the kernel has the right properties.
    assert str(k) == "(EQ() > 2) per 3"
    assert k.stationary

    # Test equality.
    assert EQ().periodic(2) == EQ().periodic(2)
    assert EQ().periodic(2) != EQ().periodic(3)
    assert Matern12().periodic(2) != EQ().periodic(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = 5 * k.stretch(5)

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Check passing in a list.
    k = EQ().periodic(np.array([1, 2]))
    k(B.randn(10, 2))

    # Check periodication of a zero.
    k = ZeroKernel()
    assert k.periodic(3) is k


def test_scaled():
    k = 2 * EQ()

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert 2 * EQ() == 2 * EQ()
    assert 2 * EQ() != 3 * EQ()
    assert 2 * EQ() != 2 * Matern12()

    # Standard tests:
    standard_kernel_tests(k)


def test_shifted():
    k = ShiftedKernel(2 * EQ(), 5)

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert Linear().shift(2) == Linear().shift(2)
    assert Linear().shift(2) != Linear().shift(3)
    assert Linear().shift(2) != DecayingKernel(1, 1).shift(2)

    # Standard tests:
    standard_kernel_tests(k)

    k = (2 * EQ()).shift(5, 6)

    # Verify that the kernel has the right properties.
    assert not k.stationary

    # Check computation.
    x1 = B.randn(10, 2)
    x2 = B.randn(5, 2)
    k = Linear()
    approx(k.shift(5)(x1, x2), k(x1 - 5, x2 - 5))

    # Check passing in a list.
    k = Linear().shift(np.array([1, 2]))
    k(B.randn(10, 2))


def test_selection():
    k = (2 * EQ().stretch(5)).select(0)

    # Verify that the kernel has the right properties.
    assert k.stationary

    # Test equality.
    assert EQ().select(0) == EQ().select(0)
    assert EQ().select(0) != EQ().select(1)
    assert EQ().select(0) != Matern12().select(0)

    # Standard tests:
    standard_kernel_tests(k)

    # Verify that the kernel has the right properties.
    k = (2 * EQ().stretch(5)).select([2, 3])
    assert k.stationary

    k = (2 * EQ().stretch(np.array([1, 2, 3]))).select([0, 2])
    assert k.stationary

    k = (2 * EQ().periodic(np.array([1, 2, 3]))).select([1, 2])
    assert k.stationary

    k = (2 * EQ().stretch(np.array([1, 2, 3]))).select([0, 2], [1, 2])
    assert not k.stationary

    k = (2 * EQ().periodic(np.array([1, 2, 3]))).select([0, 2], [1, 2])
    assert not k.stationary

    # Test computation of the kernel.
    k1 = EQ().select([1, 2])
    k2 = EQ()
    x = B.randn(10, 3)
    approx(k1(x), k2(x[:, [1, 2]]))


def test_input_transform():
    k = Linear().transform(lambda x: x - 5)

    # Verify that the kernel has the right properties.
    assert not k.stationary

    def f1(x):
        return x

    def f2(x):
        return x ** 2

    # Test equality.
    assert EQ().transform(f1) == EQ().transform(f1)
    assert EQ().transform(f1) != EQ().transform(f2)
    assert EQ().transform(f1) != Matern12().transform(f1)

    # Standard tests:
    standard_kernel_tests(k)

    # Test computation of the kernel.
    k = Linear()
    x1, x2 = B.randn(10, 2), B.randn(10, 2)

    k2 = k.transform(lambda x: x ** 2)
    k3 = k.transform(lambda x: x ** 2, lambda x: x - 5)

    approx(k(x1 ** 2, x2 ** 2), k2(x1, x2))
    approx(k(x1 ** 2, x2 - 5), k3(x1, x2))


def test_tensor_product():
    k = TensorProductKernel(lambda x: B.sum(x ** 2, axis=1))

    # Verify that the kernel has the right properties.
    assert not k.stationary

    # Test equality.
    assert k == k
    assert k != TensorProductKernel(lambda x: x)
    assert k != EQ()

    # Standard tests:
    standard_kernel_tests(k)

    # Test computation of the kernel.
    k = TensorProductKernel(lambda x: x)
    x1 = np.linspace(0, 1, 100)[:, None]
    x2 = np.linspace(0, 1, 50)[:, None]
    approx(k(x1), x1 * x1.T)
    approx(k(x1, x2), x1 * x2.T)

    k = TensorProductKernel(lambda x: x ** 2)
    approx(k(x1), x1 ** 2 * (x1 ** 2).T)
    approx(k(x1, x2), (x1 ** 2) * (x2 ** 2).T)


def test_derivative():
    k = EQ().diff(0)

    # Check that the kernel has the right properties.
    assert not k.stationary

    # Test equality.
    assert EQ().diff(0) == EQ().diff(0)
    assert EQ().diff(0) != EQ().diff(1)
    assert Matern12().diff(0) != EQ().diff(0)

    # Standard tests:
    for k in [EQ().diff(0), EQ().diff(None, 0), EQ().diff(0, None)]:
        standard_kernel_tests(k, dtype=tf.float64)

    # Check that a derivative must be specified.
    with pytest.raises(RuntimeError):
        EQ().diff(None, None)(np.array([1.0]))
    with pytest.raises(RuntimeError):
        EQ().diff(None, None).elwise(np.array([1.0]))


def test_derivative_eq():
    # Test derivative of kernel `EQ()`.
    k = EQ()
    x1 = B.randn(tf.float64, 10, 1)
    x2 = B.randn(tf.float64, 5, 1)

    # Test derivative with respect to first input.
    approx(k.diff(0, None)(x1, x2), -k(x1, x2) * (x1 - B.transpose(x2)))
    approx(k.diff(0, None)(x1), -k(x1) * (x1 - B.transpose(x1)))

    # Test derivative with respect to second input.
    approx(k.diff(None, 0)(x1, x2), -k(x1, x2) * (B.transpose(x2) - x1))
    approx(k.diff(None, 0)(x1), -k(x1) * (B.transpose(x1) - x1))

    # Test derivative with respect to both inputs.
    ref = k(x1, x2) * (1 - (x1 - B.transpose(x2)) ** 2)
    approx(k.diff(0, 0)(x1, x2), ref)
    approx(k.diff(0)(x1, x2), ref)
    ref = k(x1) * (1 - (x1 - B.transpose(x1)) ** 2)
    approx(k.diff(0, 0)(x1), ref)
    approx(k.diff(0)(x1), ref)


def test_derivative_linear():
    # Test derivative of kernel `Linear()`.
    k = Linear()
    x1 = B.randn(tf.float64, 10, 1)
    x2 = B.randn(tf.float64, 5, 1)

    # Test derivative with respect to first input.
    approx(k.diff(0, None)(x1, x2), B.ones(tf.float64, 10, 5) * B.transpose(x2))
    approx(k.diff(0, None)(x1), B.ones(tf.float64, 10, 10) * B.transpose(x1))

    # Test derivative with respect to second input.
    approx(k.diff(None, 0)(x1, x2), B.ones(tf.float64, 10, 5) * x1)
    approx(k.diff(None, 0)(x1), B.ones(tf.float64, 10, 10) * x1)

    # Test derivative with respect to both inputs.
    ref = B.ones(tf.float64, 10, 5)
    approx(k.diff(0, 0)(x1, x2), ref)
    approx(k.diff(0)(x1, x2), ref)
    ref = B.ones(tf.float64, 10, 10)
    approx(k.diff(0, 0)(x1), ref)
    approx(k.diff(0)(x1), ref)


@pytest.mark.parametrize(
    "x,result",
    [
        (np.float64([1]), np.float64([1e-20 + 1 + 1e-14])),
        (np.float32([1]), np.float32([1e-20 + 1 + 1e-7])),
    ],
)
def test_perturb(x, result):
    assert perturb(x) == result  # Test NumPy.
    assert perturb(tf.constant(x)).numpy() == result  # Test TF.


def test_perturb_type_check():
    with pytest.raises(ValueError):
        perturb(0)


@pytest.mark.parametrize("dtype", [tf.float32, tf.float64])
def test_nested_derivatives(dtype):
    x = B.randn(dtype, 10, 2)

    res = EQ().diff(0, 0).diff(0, 0)(x)
    assert ~B.isnan(res[0, 0])

    res = EQ().diff(1, 1).diff(1, 1)(x)
    assert ~B.isnan(res[0, 0])
